from numba.typed import List
#import multivelo as mv
import numpy as np

def predict_exp(tau,
                c0,
                u0,
                s0,
                alpha_c,
                alpha,
                beta,
                gamma,
                scale_cc=1,
                pred_r=True,
                chrom_open=True,
                backward=False,
                rna_only=False,
                style='multivelo'):

    if len(tau) == 0:
        return np.empty((0, 3))
    if backward:
        tau = -tau
    res = np.empty((len(tau), 3))
    if style == 'multivelo':
        
        eat = np.exp(-alpha_c * tau)
        ebt = np.exp(-beta * tau)
        egt = np.exp(-gamma * tau)
        if rna_only:
            kc = 1
            c0 = 1
        else:
            if chrom_open:
                kc = 1
            else:
                kc = 0
                alpha_c *= scale_cc

        const = (kc - c0) * alpha / (beta - alpha_c)

        res[:, 0] = kc - (kc - c0) * eat

        if pred_r:

            res[:, 1] = u0 * ebt + (alpha * kc / beta) * (1 - ebt)
            res[:, 1] += const * (ebt - eat)

            res[:, 2] = s0 * egt + (alpha * kc / gamma) * (1 - egt)
            res[:, 2] += ((beta / (gamma - beta)) *
                        ((alpha * kc / beta) - u0 - const) * (egt - ebt))
            res[:, 2] += (beta / (gamma - alpha_c)) * const * (egt - eat)

        else:
            res[:, 1] = np.zeros(len(tau))
            res[:, 2] = np.zeros(len(tau))
        return res

    elif style == 'moflow':
        if rna_only:
            kc = 1
            c0 = 1
        else:
            if chrom_open:
                kc = 1
            else:
                kc = -1
                alpha_c *= scale_cc
                
        eat = np.exp(kc * alpha_c * tau)
        ebt = np.exp(-beta * tau)
        egt = np.exp(-gamma * tau)
        

        const = alpha*c0 / (kc*alpha_c + beta)

        res[:, 0] = c0*eat

        if pred_r:

            res[:, 1] = (u0 - const)*ebt + const * eat

            res[:, 2] = alpha*const*eat/(gamma-kc*alpha_c)
            res[:, 2] += beta(u0 - const)*ebt/(gamma + beta)
            res[:, 2] += (s0 - beta*const/(gamma-kc*alpha_c) - beta*(u0 - const)/(gamma+beta))*egt
            

        else:
            res[:, 1] = np.zeros(len(tau))
            res[:, 2] = np.zeros(len(tau))
        return res
        

def generate_exp(tau_list,
                 t_sw_array,
                 alpha_c,
                 alpha,
                 beta,
                 gamma,
                 scale_cc=1,
                 model=1,
                 rna_only=False,
                 style='multivelo'):

    if beta == alpha_c:
        beta += 1e-3
    if gamma == beta or gamma == alpha_c:
        gamma += 1e-3
    switch = len(t_sw_array)
    if switch >= 1:
        tau_sw1 = np.array([t_sw_array[0]])
        if switch >= 2:
            tau_sw2 = np.array([t_sw_array[1] - t_sw_array[0]])
            if switch == 3:
                tau_sw3 = np.array([t_sw_array[2] - t_sw_array[1]])
    exp_sw1, exp_sw2, exp_sw3 = (np.empty((0, 3)),
                                 np.empty((0, 3)),
                                 np.empty((0, 3)))
    tau1 = tau_list[0]
    if switch >= 1:
        tau2 = tau_list[1]
        if switch >= 2:
            tau3 = tau_list[2]
            if switch == 3:
                tau4 = tau_list[3]
    exp1, exp2, exp3, exp4 = (np.empty((0, 3)), np.empty((0, 3)),
                              np.empty((0, 3)), np.empty((0, 3)))
    if model == 0:
        exp1 = predict_exp(tau1, 0, 0, 0, alpha_c, alpha, beta, gamma,
                           pred_r=False, scale_cc=scale_cc, rna_only=rna_only, style=style)
        if switch >= 1:
            exp_sw1 = predict_exp(tau_sw1, 0, 0, 0, alpha_c, alpha, beta,
                                  gamma, pred_r=False, scale_cc=scale_cc,
                                  rna_only=rna_only, style=style)
            exp2 = predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                               exp_sw1[0, 2], alpha_c, alpha, beta, gamma,
                               pred_r=False, chrom_open=False,
                               scale_cc=scale_cc, rna_only=rna_only, style=style)
            if switch >= 2:
                exp_sw2 = predict_exp(tau_sw2, exp_sw1[0, 0], exp_sw1[0, 1],
                                      exp_sw1[0, 2], alpha_c, alpha, beta,
                                      gamma, pred_r=False, chrom_open=False,
                                      scale_cc=scale_cc, rna_only=rna_only, style=style)
                exp3 = predict_exp(tau3, exp_sw2[0, 0], exp_sw2[0, 1],
                                   exp_sw2[0, 2], alpha_c, alpha, beta, gamma,
                                   chrom_open=False, scale_cc=scale_cc,
                                   rna_only=rna_only, style=style)
                if switch == 3:
                    exp_sw3 = predict_exp(tau_sw3, exp_sw2[0, 0],
                                          exp_sw2[0, 1], exp_sw2[0, 2],
                                          alpha_c, alpha, beta, gamma,
                                          chrom_open=False, scale_cc=scale_cc,
                                          rna_only=rna_only, style=style)
                    exp4 = predict_exp(tau4, exp_sw3[0, 0], exp_sw3[0, 1],
                                       exp_sw3[0, 2], alpha_c, 0, beta, gamma,
                                       chrom_open=False, scale_cc=scale_cc,
                                       rna_only=rna_only, style=style)
    elif model == 1:
        exp1 = predict_exp(tau1, 0, 0, 0, alpha_c, alpha, beta, gamma,
                           pred_r=False, scale_cc=scale_cc, rna_only=rna_only, style=style)
        if switch >= 1:
            exp_sw1 = predict_exp(tau_sw1, 0, 0, 0, alpha_c, alpha, beta,
                                  gamma, pred_r=False, scale_cc=scale_cc,
                                  rna_only=rna_only, style=style)
            exp2 = predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                               exp_sw1[0, 2], alpha_c, alpha, beta, gamma,
                               scale_cc=scale_cc, rna_only=rna_only, style=style)
            if switch >= 2:
                exp_sw2 = predict_exp(tau_sw2, exp_sw1[0, 0], exp_sw1[0, 1],
                                      exp_sw1[0, 2], alpha_c, alpha, beta,
                                      gamma, scale_cc=scale_cc,
                                      rna_only=rna_only, style=style)
                exp3 = predict_exp(tau3, exp_sw2[0, 0], exp_sw2[0, 1],
                                   exp_sw2[0, 2], alpha_c, alpha, beta, gamma,
                                   chrom_open=False, scale_cc=scale_cc,
                                   rna_only=rna_only, style=style)
                if switch == 3:
                    exp_sw3 = predict_exp(tau_sw3, exp_sw2[0, 0],
                                          exp_sw2[0, 1], exp_sw2[0, 2],
                                          alpha_c, alpha, beta, gamma,
                                          chrom_open=False, scale_cc=scale_cc,
                                          rna_only=rna_only, style=style)
                    exp4 = predict_exp(tau4, exp_sw3[0, 0], exp_sw3[0, 1],
                                       exp_sw3[0, 2], alpha_c, 0, beta, gamma,
                                       chrom_open=False, scale_cc=scale_cc,
                                       rna_only=rna_only, style=style)
    elif model == 2:
        exp1 = predict_exp(tau1, 0, 0, 0, alpha_c, alpha, beta, gamma,
                           pred_r=False, scale_cc=scale_cc, rna_only=rna_only, style=style)
        if switch >= 1:
            exp_sw1 = predict_exp(tau_sw1, 0, 0, 0, alpha_c, alpha, beta,
                                  gamma, pred_r=False, scale_cc=scale_cc,
                                  rna_only=rna_only, style=style)
            exp2 = predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                               exp_sw1[0, 2], alpha_c, alpha, beta, gamma,
                               scale_cc=scale_cc, rna_only=rna_only, style=style)
            if switch >= 2:
                exp_sw2 = predict_exp(tau_sw2, exp_sw1[0, 0], exp_sw1[0, 1],
                                      exp_sw1[0, 2], alpha_c, alpha, beta,
                                      gamma, scale_cc=scale_cc,
                                      rna_only=rna_only, style=style)
                exp3 = predict_exp(tau3, exp_sw2[0, 0], exp_sw2[0, 1],
                                   exp_sw2[0, 2], alpha_c, 0, beta, gamma,
                                   scale_cc=scale_cc, rna_only=rna_only, style=style)
                if switch == 3:
                    exp_sw3 = predict_exp(tau_sw3, exp_sw2[0, 0],
                                          exp_sw2[0, 1], exp_sw2[0, 2],
                                          alpha_c, 0, beta, gamma,
                                          scale_cc=scale_cc, rna_only=rna_only, style=style)
                    exp4 = predict_exp(tau4, exp_sw3[0, 0], exp_sw3[0, 1],
                                       exp_sw3[0, 2], alpha_c, 0, beta, gamma,
                                       chrom_open=False, scale_cc=scale_cc,
                                       rna_only=rna_only, style=style)
    return (exp1, exp2, exp3, exp4), (exp_sw1, exp_sw2, exp_sw3)


def simulate(n_cell, param, t_sw_array, total_h=20, direction='complete',
            model=1, style='multivelo'):
    
    t = np.linspace(0, total_h, n_cell)

    param = np.clip(param, 1e-3, None)
    if len(param.shape) == 1:
        
        alpha_c, alpha, beta, gamma = param
        if beta == alpha_c:
            beta += 1e-3
        if gamma == beta or gamma == alpha_c:
            gamma += 1e-3
    else:
        alpha_c, alpha, beta, gamma = [param[:, i] for i in range(4)]    
    if direction=='complete':
        state0 = t <= t_sw_array[0]
        state1 = (t_sw_array[0] < t) & (t <= t_sw_array[1])
        state2 = (t_sw_array[1] < t) & (t <= t_sw_array[2])
        state3 = t_sw_array[2] < t

        tau1 = t[state0]
        tau2 = t[state1] - t_sw_array[0]
        tau3 = t[state2] - t_sw_array[1]
        tau4 = t[state3] - t_sw_array[2]
        tau_list = [tau1, tau2, tau3, tau4]
        switch = np.sum(t_sw_array < total_h)
        typed_tau_list = List()
        [typed_tau_list.append(x) for x in tau_list]

        exp_list, exp_sw_list = generate_exp(typed_tau_list,
                        t_sw_array[:switch],
                        alpha_c,
                        alpha,
                        beta,
                        gamma,
                        scale_cc=1,
                        model=model,
                        rna_only=False,
                        style=style)
        c = np.empty(len(t))
        u = np.empty(len(t))
        s = np.empty(len(t))
        for i, ii in enumerate([state0, state1, state2, state3]):
            if np.any(ii):
                c[ii] = exp_list[i][:, 0]
                u[ii] = exp_list[i][:, 1]
                s[ii] = exp_list[i][:, 2]
    elif direction=='on':
        state0 = t <= t_sw_array[0]
        state1 = t > t_sw_array[0]
        tau1 = t[state0]
        tau2 = t[state1] - t_sw_array[0]
        tau_sw1 = np.array([t_sw_array[0]])
        
        exp_sw1 = np.empty((0, 3))
        exp_sw1 = mv.dynamical_chrom_func.predict_exp(tau_sw1, 0, 0, 0, alpha_c, alpha, beta,
                                    gamma, pred_r=False)
        exp1 = mv.dynamical_chrom_func.predict_exp(tau1, 0, 0, 0, alpha_c, alpha, beta, gamma,
                        pred_r=False)
        exp2 = mv.dynamical_chrom_func.predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                            exp_sw1[0, 2], alpha_c, alpha, beta, gamma,
        )
        exp_list = [exp1, exp2]
        c = np.empty(len(t))
        u = np.empty(len(t))
        s = np.empty(len(t))
        for i, ii in enumerate([state0, state1]):
            if np.any(ii):
                c[ii] = exp_list[i][:, 0]
                u[ii] = exp_list[i][:, 1]
                s[ii] = exp_list[i][:, 2]
        
    else:
        state0 = t <= t_sw_array[0]
        state1 = t > t_sw_array[0]
        tau1 = t[state0]
        tau2 = t[state1] - t_sw_array[0]
        tau_sw1 = np.array([t_sw_array[0]])
        
        exp_sw1 = np.empty((0, 3))
        if model==1:
            exp_sw1 = mv.dynamical_chrom_func.predict_exp(tau_sw1, 1, 1, 1, alpha_c, alpha, beta,
                                    gamma, chrom_open=False)
            exp1 = mv.dynamical_chrom_func.predict_exp(tau1,  1, 1, 1, alpha_c, alpha, beta, gamma,
                        chrom_open=False)
            exp2 = mv.dynamical_chrom_func.predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                            exp_sw1[0, 2], alpha_c, 0, beta, gamma,
                            chrom_open=False,
            )
            exp_list = [exp1, exp2]
        elif model==2:
            exp_sw1 = mv.dynamical_chrom_func.predict_exp(tau_sw1, 1, 1, 1, alpha_c, 0, beta,
                                    gamma)
            exp1 = mv.dynamical_chrom_func.predict_exp(tau1,  1, 1, 1,  alpha_c, 0, beta, gamma,
                        )
            exp2 = mv.dynamical_chrom_func.predict_exp(tau2, exp_sw1[0, 0], exp_sw1[0, 1],
                            exp_sw1[0, 2], alpha_c, 0, beta, gamma,
                            chrom_open=False, 
            )
            exp_list = [exp1, exp2]
            
        c = np.empty(len(t))
        u = np.empty(len(t))
        s = np.empty(len(t))
        for i, ii in enumerate([state0, state1]):
            if np.any(ii):
                c[ii] = exp_list[i][:, 0]
                u[ii] = exp_list[i][:, 1]
                s[ii] = exp_list[i][:, 2]


    
    return c, u, s

    
    


import numpy as np
from scipy.ndimage import uniform_filter1d
from fastdtw import fastdtw
import matplotlib.pyplot as plt


def normalize(x):
    return (x - np.min(x)) / (np.max(x) - np.min(x))
def smooth(x, size=3):# smoothing
    return uniform_filter1d(x, size=size)
def pad_ends(x):# Add zero to both ends
    return np.concatenate([[0], x, [0]])

def bin_and_average(x, time, n_bins=20):
    bins = np.linspace(np.min(time), np.max(time), n_bins + 1)
    binned = np.digitize(time, bins) - 1
    out = np.array([np.mean(x[binned == i]) for i in range(n_bins)])
    return out

def get_dtw(adata, gene, genetime=False, 
            timekey = 'velo_s_pseudotime', n_bins=20):
    """DTW analysis
        
    Parameters
    ---------
    adata: `anndata.AnnData`
        anndata that contains the chromatin accessibility, unspliced abundance, spliced abundance, embedding space, and pseudotime information. 
    gene_time: optional, `bool` (default: False)
        `True` if use gene specific time. `False` if use global time.
    time_key: optional, `str` (default: velo_s_pseudotime)
        obs or layer key to annotate pseudotime.
    n_bins: optional, `int` (default: 20)
        the number of bin to discritize time
    Returns
    -------
    time_pad: `numpy.ndarray`
        A linearly spaced array representing the padded time points after dynamic time warping (DTW).
    c_pad: `numpy.ndarray`
        Padded and smoothed chromatin accessibility values for the gene over time.
    u_pad: `numpy.ndarray`
        Padded and smoothed unspliced abundance values for the gene over time.
    s_pad: `numpy.ndarray`
        Padded and smoothed spliced abundance values for the gene over time.
    path_c_s: `list of tuples`
        The optimal alignment path (DTW path) between chromatin accessibility (c_pad) and spliced abundance (s_pad).
    path_u_s: `list of tuples`
        The optimal alignment path (DTW path) between unspliced abundance (u_pad) and spliced abundance (s_pad).
    time_lag_c_s: `list of float`
        The time lags between the aligned chromatin accessibility and spliced abundance based on the DTW path.
    time_lag_u_s: `list of float`
        The time lags between the aligned unspliced abundance and spliced abundance based on the DTW path.
        """
    
    if genetime:
        time = adata[:, gene].layers[timekey].ravel()
    else:
        time = adata.obs[timekey].values
    time = np.clip(time, 0, np.quantile(time, 0.95))

    c = adata[:, gene].layers['Mc'].ravel()
    u = adata[:, gene].layers['Mu'].ravel()
    s = adata[:, gene].layers['Ms'].ravel()
    
    c = bin_and_average(c, time, n_bins)
    u = bin_and_average(u, time, n_bins)
    s = bin_and_average(s, time, n_bins)


    c_norm = normalize(c)
    u_norm = normalize(u)
    s_norm = normalize(s)

    c_smooth = smooth(c_norm)
    u_smooth = smooth(u_norm)
    s_smooth = smooth(s_norm)

    c_pad = pad_ends(c_smooth)
    u_pad = pad_ends(u_smooth)
    s_pad = pad_ends(s_smooth)
    
    time_pad = np.linspace(0, 1, len(c_pad))


    distance_c_s, path_c_s = fastdtw(c_pad, s_pad, )
    distance_u_s, path_u_s = fastdtw(u_pad, s_pad, )

    time_lag_c_s = [time_pad[j] - time_pad[i] for i, j in path_c_s]
    time_lag_u_s = [time_pad[j] - time_pad[i] for i, j in path_u_s]
    return(time_pad, c_pad, u_pad, s_pad, path_c_s, path_u_s, time_lag_c_s, time_lag_u_s)


def plot_dtw(adata, gene, genetime=False, timekey = 'velo_s_pseudotime', n_bins=20,
         figsave=None):
    ):
    """
    Plot the DTW alignment of chromatin accessibility, unspliced, and spliced abundance profiles over time 
    along with the corresponding time lags.

    Parameters
    ----------
    adata : `anndata.AnnData`
        Annotated data object containing the chromatin accessibility, unspliced and spliced abundance,
        embedding space, and pseudotime information.
    
    gene : `str`
        The gene name to plot the DTW alignment for.
    
    genetime : `bool`, optional (default: False)
        If `True`, use gene-specific time values. If `False`, use the global pseudotime.
    
    timekey : `str`, optional (default: 'velo_s_pseudotime')
        The key in `adata.obs` or `adata.layers` that stores the pseudotime information.
    
    n_bins : `int`, optional (default: 20)
        The number of bins to discretize the time values for averaging.
    
    figsave : `str` or `None`, optional (default: None)
        If provided, the figure will be saved to the specified file path. If `None`, the figure will not be saved.

    Returns
    -------
    None
    """
    time_pad, c_pad, u_pad, s_pad, path_c_s, path_u_s, time_lag_c_s, time_lag_u_s = get_dtw(adata, gene, genetime, timekey, n_bins)

    fig, axs = plt.subplots(2, 1, figsize=(4.8, 3.2), sharex=True)

    # Top: c vs s
    axs[0].plot(time_pad, c_pad, label='c', color='#102E50')
    axs[0].plot(time_pad, u_pad, label='u', color='#F5C45E')
    axs[0].plot(time_pad, s_pad, label='s', color='#BE3D2A')
    
    for i, j in path_c_s[::3]:  # skip some lines for clarity
        axs[0].plot([time_pad[i], time_pad[j]], [c_pad[i], s_pad[j]], 'k--', linewidth=0.5)
    for i, j in path_u_s[::3]:
        axs[0].plot([time_pad[i], time_pad[j]], [u_pad[i], s_pad[j]], 'k--', linewidth=0.5)
    
    axs[0].set_ylabel("Norm values")
    axs[0].legend()
    axs[0].set_title("DTW Alignment")


    axs[1].plot([time_pad[i] for i, _ in path_c_s], time_lag_c_s, label='c-s lag', color='#3E3F5B')
    axs[1].plot([time_pad[i] for i, _ in path_u_s], time_lag_u_s, label='u-s lag', color='#8AB2A6')
    axs[1].axhline(0, color='gray', linestyle='--')
    axs[1].set_ylabel("Time lag")
    axs[1].set_xlabel("Time")
    axs[1].legend()

    
    fig.suptitle(gene)

    plt.tight_layout()
    plt.show()
    
    if figsave is not None:
        fig.savefig(figsave)



import matplotlib.pyplot as plt
import numpy as np


def scatter_cell(adata,
                 gene_list,
                 by='umap',
                 color_by='s',
                 cmap='coolwarm',
                 vmax=None,
                 vmin=None,
                 center=None,
                 n_cols=5,
                 axis_on=True,
                 frame_on=True,
                 figsize=None,
                 pointsize=2,
                 alpha=.5,
                 color_bar=False,
                 **kwargs
                 ):
    """Plot the kinetic parameters ('alpha_c', 'alpha', 'beta', 'gamma', 'splice', 'unsplice', or 'pseudotime') of one gene on the embedding space.
        
    Arguments
    ---------
    adata: :class:`~anndata.AnnData` or `~mudata.MuData`
        Data result from dynamics recovery.
    gene_list: `str`,  list of `str`
        List of gene_list to plot.
    by: `str` (default: `s`)
    
    """  
    
    from pandas.api.types import is_numeric_dtype, is_categorical_dtype
    
    if color_by in adata.obs and is_numeric_dtype(adata.obs[color_by]):
        types = None
        colors = adata.obs[color_by].values
    elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]) \
            and color_by+'_colors' in adata.uns.keys():
        types = adata.obs[color_by].cat.categories
        colors = adata.uns[f'{color_by}_colors']
    elif color_by in adata.layers and is_numeric_dtype(adata.layers[color_by]):
        types = None
        colors=None
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata '
                         'obs, and the colors of categories can be found in '
                         'adata uns.')
        
        
    gene_list = np.array(gene_list)
    missing_genes = gene_list[~np.isin(gene_list, adata.var_names)]

    gene_list = gene_list[np.isin(gene_list, adata.var_names)]
    gn = len(gene_list)
    if gn == 0:
        return
    elif gn < n_cols:
        n_cols = gn
        
    fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3.2*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize)
    fig.patch.set_facecolor('white')
    
    embed = adata.obsm[f'X_{by}']
    x = embed[:, 0]
    y = embed[:, 1]
    
    for g, gene in enumerate(gene_list):
        row = g // n_cols
        col = g % n_cols
        ax = axs[row, col]
        
        if types is not None:
            for i in range(len(types)):
                filt = adata.obs[color_by] == types[i]
                filt = np.arange(adata.n_obs)[np.ravel(filt)]
                if type(colors) == dict:
                    color = colors[types[i]]
                
                ax.scatter(x[filt], y[filt], s=pointsize, c=color, alpha=alpha,
                 **kwargs)
        else:
            if colors is not None:
                im = ax.scatter(x, y,  s=pointsize, c=colors, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            else:
                color = adata[:, gene].layers[color_by]
                if center is not None:
                    vmax = max(abs(color))
                    vmin = -vmax
                im = ax.scatter(x, y, s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            if color_bar:
                plt.colorbar(im, ax=ax)
                
        title = gene
        ax.set_title(f'{title}', fontsize=11)
        
        if not frame_on:
            ax.xaxis.set_ticks_position('none')
            ax.yaxis.set_ticks_position('none')
            ax.set_frame_on(False)
    fig.tight_layout()
    return fig, axs

def scatter_cell_multi(adata,
                 gene,
                 by='umap',
                 color_by='expression',
                 vmax=None,
                 vmin=None,
                 center=None,
                 n_cols=5,
                 axis_on=True,
                 frame_on=True,
                 figsize=None,
                 pointsize=2,
                 alpha=.5,
                 cmap='coolwarm',
                 **kwargs
                 ):
    """Plot the kinetic parameters ('alpha', 'beta', 'gamma', 'splice', 'unsplice', or 'pseudotime') of one gene on the embedding space.
        
    Arguments
    ---------
    adata: :class:`~anndata.AnnData` or `~mudata.MuData`
        Data result from dynamics recovery.
    gene_list: `str`,  list of `str`
        List of gene_list to plot.
    by: `str` (default: `s`)
    
    """  
    

    if gene not in adata.var_names:
        raise ValueError()

    if color_by == 'expression':
        fig, axs = plt.subplots(1, 3, squeeze=False,
                                figsize=(3.2*3, 2.7)
                                if figsize is None else figsize)
    elif color_by == 'velocity':
        fig, axs = plt.subplots(1, 3, squeeze=False,
                                figsize=(3.2*3, 2.7)
                                if figsize is None else figsize)
    elif color_by =='parameter':
        fig, axs = plt.subplots(1, 4, squeeze=False,
                                figsize=(3.2*4, 2.7)
                                if figsize is None else figsize)
    elif color_by == 'all':
        fig, axs = plt.subplots(3, 4, squeeze=False,
                                figsize=(3.2*4, 2.7*3)
                                if figsize is None else figsize)
    else:
        raise ValueError('color key must be the one of expression,' 
                         'velocity, parameter, or all')

    fig.patch.set_facecolor('white')
    
    embed = adata.obsm[f'X_{by}']
    x = embed[:, 0]
    y = embed[:, 1]
    
    if color_by == 'expression':
        if vmax is None:
            vmax = 1
        if vmin is None:
            vmin = 0
        for i, c in zip(range(3), ['c', 'u', 's']):
            ax = axs[0, i]
            color = adata[:, gene].layers[c]
            
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
    elif color_by == 'velocity':
        if center is None:
            center = 0

        for i, c in zip(range(3), ['velo_c', 'velo_u', 'velo_s']):
            ax = axs[0, i]
            color = adata[:, gene].layers[c]
            vmax = max(abs(color))
            vmin = -vmax
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
    elif color_by == 'parameter':
        for i, c in enumerate(['alpha_c', 'alpha', 'beta', 'gamma']):
            ax = axs[0, i]
            color = adata[:, gene].layers[c]
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
    else:
        for i, c in zip(range(3), ['c', 'u', 's']):
            ax = axs[0, i]
            color = adata[:, gene].layers[c]
            vmax = 1
            vmin = 0
            cmap = 'coolwarm'
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
        for i, c in zip(range(3), ['velo_c', 'velo_u', 'velo_s']):
            ax = axs[1, i]
            color = adata[:, gene].layers[c]
            vmax = max(abs(color))
            vmin = -vmax
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       vmax=vmax, vmin=vmin, cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
        for i, c in zip(range(4), ['alpha_c', 'alpha', 'beta', 'gamma']):
            ax = axs[2, i]
            color = adata[:, gene].layers[c]
            cmap = 'gnuplot'
            im = ax.scatter(x, y,  s=pointsize, c=color, alpha=alpha,
                       cmap=cmap, **kwargs)
            ax.set_title(c)
            plt.colorbar(im, ax=ax)
    for ax in axs:
        if not frame_on:
            ax.xaxis.set_ticks_position('none')
            ax.yaxis.set_ticks_position('none')
            ax.set_frame_on(False)
    fig.tight_layout()
    return fig, axs
            


# Portions of this code are adapted from the cellDancer repository:
# https://github.com/GuangyuWangLab2021/cellDancer/

import os
import glob
import pandas as pd
import numpy as np
import random
from scipy import sparse
from sklearn.metrics import pairwise_distances
from sklearn.mixture import GaussianMixture
from scipy.spatial.distance import cdist


import torch
import torch.nn as nn
import torch.nn.functional as F


import pytorch_lightning as pl
from torch.utils.data import *
from pytorch_lightning.callbacks.early_stopping import EarlyStopping

from sklearn.neighbors import NearestNeighbors
from joblib import Parallel, delayed
from tqdm import tqdm
import pkg_resources
import warnings

warnings.filterwarnings("ignore", category=DeprecationWarning)
warnings.simplefilter("ignore", UserWarning)

import logging
logging.getLogger("pytorch_lightning").setLevel(logging.ERROR)

from multiprocessing import Manager

from .sampling import *

def cosine_similarity(cnv, unv, snv, vc, vu, vs, rna_only=False):
    """Cost function
    Return:
        list of cosine distance and a list of the index of the next cell
    """
    if rna_only:
        nv = torch.stack([unv, snv], dim=-1)
        v = torch.stack([vu, vs], dim=-1)

        cosine = F.cosine_similarity(nv, v, dim=-1)

        cosine_max, cosine_max_idx = torch.max(cosine, dim=0)
        return 1 - cosine_max , cosine_max_idx


    else:
        nv = torch.stack([cnv, unv, snv], dim=-1)
        v = torch.stack([vc, vu, vs], dim=-1)

        cosine = F.cosine_similarity(nv, v, dim=-1)

        cosine_max, cosine_max_idx = torch.max(cosine, dim=0)
        return 1 - cosine_max, cosine_max_idx



def get_loss(c, u, s, param, indices, 
             direction=None, rna_only=False, warm=False
             ):
    
    if rna_only:
        alpha, beta, gamma = param
        n = c.shape[0]
        
        vu = alpha - beta*u
        vs = beta*u - gamma*s
        
        indices_ = indices[:, 1:].T
        
        cnv, unv, snv = c[indices_] - c, u[indices_] - u, s[indices_] - s   
        cosine_, cosine_idx= cosine_similarity(cnv, unv, snv, vc, vu, vs, rna_only=True)
        
        velo = [vc, vu, vs]
        loss = torch.mean(cosine_)
        
        stateon = (vc>=0).type(torch.float32)
        stateoff = (vc<0).type(torch.float32)
        states = [stateon, stateoff]

        return loss, velo, param, states
    else:
        alpha_c, alpha, beta, gamma = param

        n = c.shape[0]
        vcopen = alpha_c*(1-c)
        vcclose = alpha_c*(-c)

        vu = alpha*c - beta*u
        vs = beta*u - gamma*s
        indices_ = indices[:, 1:].T

        cnv, unv, snv = c[indices_] - c, u[indices_] - u, s[indices_] - s   
        

        vc = torch.stack([vcopen, vcclose], dim=1)

        if warm:
            if direction=='on':
                vu = torch.clamp(vu, 0., None)
                vs = torch.clamp(vs, 0., None)
            elif direction=='off':
                vu = torch.clamp(vu, None, 0.)
                vs = torch.clamp(vs, None, 0.)
            else:
                _ = (u<s)
                vu[_] = torch.clamp(vu[_], None, 0.)
                vs[_] = torch.clamp(vs[_], None, 0.)
                vu[~_] = torch.clamp(vu[~_], 0., None)
                vs[~_] = torch.clamp(vs[~_], 0., None)
                
            cosine_, cosine_idx= cosine_similarity(cnv, unv, snv, vc, vu, vs, rna_only=True)
            cnv_, unv_, snv_ = cnv[cosine_idx, torch.arange(n)], unv[cosine_idx, torch.arange(n)], snv[cosine_idx, torch.arange(n)]
            nv = torch.stack([cnv_, unv_, snv_], dim=-1)
            
            vopen = torch.stack([vcopen, vu, vs], dim=-1)
            vclose = torch.stack([vcclose, vu, vs], dim=-1)

            cosine1 = 1-F.cosine_similarity(nv, vopen, dim=-1)
            cosine2 = 1-F.cosine_similarity(nv, vclose, dim=-1)

            cosineloss = torch.stack([cosine1, cosine2], dim=-1)
            cosineloss, _ = torch.min(cosineloss, dim=1) 

            vc = vc[torch.arange(n), _]
            loss = torch.mean(cosine1 + cosine2)/2 #choice between vc open or close is not crucial in warmup phase
            
        else:
            cosine_, cosine_idx= cosine_similarity(cnv, unv, snv, vc, vu, vs, rna_only=True)
            cnv_, unv_, snv_ = cnv[cosine_idx, torch.arange(n)], unv[cosine_idx, torch.arange(n)], snv[cosine_idx, torch.arange(n)]
            nv = torch.stack([cnv_, unv_, snv_], dim=-1)
            vopen = torch.stack([vcopen, vu, vs], dim=-1)
            vclose = torch.stack([vcclose, vu, vs], dim=-1)

            cosine1 = 1-F.cosine_similarity(nv, vopen, dim=-1)
            cosine2 = 1-F.cosine_similarity(nv, vclose, dim=-1)

            cosineloss = torch.stack([cosine1, cosine2], dim=-1)

            cosineloss, _ = torch.min(cosineloss, dim=1)

            vc = vc[torch.arange(n), _]
            loss = torch.mean(cosineloss)

        param = [alpha_c, alpha, beta, gamma] 

        vc = vc[indices_]
        vc = torch.mean(vc, dim=0)
        velo = [vc, vu, vs]

        stateon = (vc>=0).type(torch.float32)
        stateoff = (vc<0).type(torch.float32)
        states = [stateon, stateoff]
        
        return loss, velo, param, states, direction


class DNN_layer(nn.Module):

    """Define network structure.
    """

    def __init__(self, rna_only=False, 
                 direction_='complete',
                 warmup=20):
        super().__init__()
        self.rna_only = rna_only
        self.direction = direction_
        self.warmup=warmup
        
        if self.rna_only:
            self.l1 = nn.Linear(2, 100)
            self.l2 = nn.Linear(100, 100)
            self.l3 = nn.Linear(100, 3)
        
        else:
            
            self.param = nn.Sequential(
                nn.Linear(3, 64),
                nn.ReLU(),
                nn.Linear(64, 48),
                nn.ReLU(),
                nn.Linear(48, 32),
                nn.ReLU(),
                nn.Linear(32, 3),
                nn.Sigmoid()
            )
            
            self.param_c = nn.Sequential(
                nn.Linear(3, 64),
                nn.ReLU(),
                nn.Linear(64, 48),
                nn.ReLU(),
                nn.Linear(48, 32),
                nn.ReLU(),
                nn.Linear(32, 1),
                nn.Sigmoid()
            )
        if self.direction == 'on':
            self.alpha0 = nn.Parameter(torch.tensor(2.0))
            self.gamma0 = nn.Parameter(torch.tensor(1.0))
            
        else:
            self.alpha0 = nn.Parameter(torch.tensor(1.0))
            self.gamma0 = nn.Parameter(torch.tensor(1.0))
            

    def forward(self, c, u, s, indices, epoch=None):
        if self.rna_only:
            
            input = torch.stack([u, s], dim=1)
            
            x = self.l1(input)
            x = F.leaky_relu(x)
            x = self.l2(x)
            x = F.leaky_relu(x)
            x = self.l3(x)
            output = torch.sigmoid(x)
            alpha = output[:,0]
            beta = output[:,1]
            gamma = output[:,2]
            c = torch.ones_like(u)
            
        else:
            n = c.shape[0]
                    
            input = torch.stack([c, u, s], dim=1)
            
            output= self.param(input)
            self.alpha0.data.clamp_(1., 10.)
            self.gamma0.data.clamp_(.5, 1.)
            alpha = output[:,0]*self.alpha0
            beta = output[:,1]
            gamma = output[:,2]*self.gamma0

            alpha_c = self.param_c(input)[:, 0]

            param = [alpha_c, alpha, beta, gamma]
            
            if epoch > self.warmup:
                warm=False
            else:
                warm=True
            
            loss, velo, param, states, direction = get_loss(
                c, u, s, param, indices, self.direction, self.rna_only, warm )
            vc, vu, vs = velo
            alpha_c, alpha, beta, gamma = param

        
        return loss, vc, vu, vs, alpha_c, alpha, beta, gamma, states
          

            
class DNN_module(nn.Module):
    '''
    calculate loss function
    load network "DNN_layer"
    predict splice_predict and unsplice_predict
    '''
    def __init__(self, module, n_neighbors = None, rna_only=False, device=None):
        super().__init__()
        self.module = module
        self.n_neighbors = n_neighbors
        self.rna_only=rna_only
        if device is None:
            device = torch.device('cpu')
        else:
            self.device = device

    def velocity_calculate(self, 
                           c,
                           u,
                           s,
                           embedding1,
                           embedding2, 
                        epoch=None,
                        test=False):
        '''
        add embedding
        for real dataset
        calculate loss function
        predict unsplice_predict splice_predict from network 
        '''
        
        #generate neighbor indices and expr dataframe

        
        if self.rna_only:
            points = np.array([u.cpu().numpy(), s.cpu().numpy()]).transpose()
            distance_matrix = cdist(points, points, metric='mahalanobis')
            indices = np.argsort(distance_matrix, axis=1)[:, :self.n_neighbors]
            
        else:
            points = np.array([c.cpu().numpy(), u.cpu().numpy(), s.cpu().numpy()]).transpose()
            distance_matrix = cdist(points, points, metric='mahalanobis')
            indices = np.argsort(distance_matrix, axis=1)[:, :self.n_neighbors]

        loss, vc, vu, vs, alpha_c, alpha, beta, gamma, states = self.module(
            c, u, s, indices, epoch)
        
        if torch.isnan(vc).any():
            print('vc has nan')
        if torch.isinf(vc).any():
            print('vc has inf')
        
        return loss, vc, vu, vs, alpha_c, alpha, beta, gamma, states


    def summary_para_validation(self, cost_mean): 
        loss_df = pd.DataFrame({'cost': cost_mean}, index=[0])
        return(loss_df)


    def summary_para(self, c, u, s,
                     vc, vu, vs,
                    alpha_c,
                    alpha, beta, gamma, 
                     stateon, stateoff
                    ): 
        output_df = pd.merge(pd.DataFrame(c, columns=['c']),pd.DataFrame(u, columns=['u']), left_index=True, right_index=True) 
        
        output_df['s'] = s
        output_df['vc'] = vc
        output_df['vu'] = vu
        output_df['vs'] = vs
        output_df['alpha_c'] = alpha_c
        output_df['alpha'] = alpha
        output_df['beta'] = beta
        output_df['gamma'] = gamma
        output_df['stateon'] = stateon
        output_df['stateoff'] = stateoff

        return output_df



class ltModule(pl.LightningModule):
    '''
    train network using "DNN_module"
    '''
    def __init__(self, 
                backbone=None, 
                learning_rate=None,
                optimizer='Adam',
                cost_type='smooth',
                average_cost_window_size=10,
                smooth_weight=0.9
                ):
        super().__init__()
        self.backbone = backbone
        self.validation_loss = []
        self.learning_rate=learning_rate
        self.optimizer=optimizer
        self.save_hyperparameters()
        self.get_loss=1000
        self.cost_type=cost_type
        self.average_cost_window_size=average_cost_window_size # will be used only when cost_tpye.isin(['average', 'median'])
        self.cost_window=[]
        self.smooth_weight=smooth_weight
        
    def save(self, model_path):
        self.backbone.module.save(model_path)    # save network

    def configure_optimizers(self):     # define optimizer
        if self.optimizer=="Adam":
            optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate, betas=(0.9, 0.999), eps=10**(-8), weight_decay=0.004, amsgrad=False)

        elif self.optimizer=="SGD":
            optimizer = torch.optim.SGD(self.parameters(), lr=self.learning_rate, momentum=0.8)


        return optimizer

    def training_step(self, batch, batch_idx):
        '''
        traning network
        batch: [] output returned from realDataset.__getitem__
        
        '''

        cs, us, ss, embedding1s, embedding2s, gene_names, _ = batch
        c, u, s, embedding1, embedding2, gene_name= cs[0], us[0], ss[0], embedding1s[0], embedding2s[0], gene_names[0]
        
        cost, vc, vu, vs, alpha_c, alpha, beta, gamma, states = self.backbone.velocity_calculate( \
                c, u, s, embedding1, embedding2, self.current_epoch)

        if self.cost_type=='average': # keep the window len <= check_val_every_n_epoch
            if len(self.cost_window)<self.average_cost_window_size:
                self.cost_window.append(cost)
            else:
                self.cost_window.pop(0)
                self.cost_window.append(cost)
            self.get_loss = torch.mean(torch.stack(self.cost_window))
            self.log("loss", self.get_loss)
            
        elif self.cost_type=='median': # keep the window len <= check_val_every_n_epoch
            if len(self.cost_window)<self.average_cost_window_size:
                self.cost_window.append(cost)
            else:
                self.cost_window.pop(0)
                self.cost_window.append(cost)
            self.get_loss = torch.median(torch.stack(self.cost_window))
            self.log("loss", self.get_loss)
            
        elif self.cost_type=='smooth':
            if self.get_loss==1000:
                self.get_loss=cost
            smoothed_val = cost * self.smooth_weight + (1 - self.smooth_weight) * self.get_loss  # calculate smoothed value
            self.get_loss = smoothed_val  
            self.log("loss", self.get_loss)
        else:
            self.get_loss = cost
            self.log("loss", self.get_loss) 

        return cost


    def validation_step(self, batch, batch_idx):
        '''
        predict unsplice_predict, splice_predict on the training dataset
        '''
        if self.current_epoch!=0:
            cost = self.get_loss.data.cpu().numpy()
            self.validation_loss.append(cost)

    def test_step(self, batch, batch_idx):
        cs, us, ss, embedding1s, embedding2s, gene_names, _ = batch
        c, u, s, embedding1, embedding2, gene_name= cs[0], us[0], ss[0], embedding1s[0], embedding2s[0], gene_names[0]
        
        cost, vc, vu, vs, alpha_c, alpha, beta, gamma, states = \
        self.backbone.velocity_calculate(
            c, u, s, 
            embedding1, embedding2, self.current_epoch, test=True)
        
        stateon, stateoff= states
        self.test_df= self.backbone.summary_para(
            c.data.cpu().numpy(), u.data.cpu().numpy(), s.data.cpu().numpy(),
            vc.data.cpu().numpy(), vu.data.cpu().numpy(), vs.data.cpu().numpy(),
                    alpha_c.data.cpu().numpy(), 
                    alpha.data.cpu().numpy(), beta.data.cpu().numpy(), gamma.data.cpu().numpy(), 
            stateon.data.cpu().numpy(), stateoff.data.cpu().numpy(),
        )
        self.test_df.insert(0, "gene_name", gene_name)
        self.test_df.insert(0, "cellIndex", self.test_df.index)
        self.test_df.insert(len(self.test_df.columns), "epoch", self.current_epoch)
        self.test_df.insert(len(self.test_df.columns), "loss", cost.item())

class getItem(Dataset): 
    '''
    load training and test data
    '''
    def __init__(self, c, u, s, embedding,
                 rna_only=False,
                 data_fit=None, data_predict=None, 
                 datastatus="predict_dataset", gene_name=None,
                 permutation_ratio=0.1, norm_data=True, norm_cell_distribution=False): 
        self.c = c
        self.u = u
        self.s = s
        self.embedding = embedding
        self.data_fit=data_fit
        self.data_predict=data_predict
        self.rna_only=rna_only
        self.datastatus=datastatus
        self.permutation_ratio=permutation_ratio
        self.norm_data=norm_data
        self.norm_max_unsplice=None
        self.norm_max_splice=None
        self.norm_cell_distribution=norm_cell_distribution
        self.gene_name = gene_name
        

        
        gpdist = pairwise_distances(
                embedding)
        subset_cells = s > 0.1*np.percentile(s, 99)
        subset_cells = np.where(subset_cells)[0]
        if len(subset_cells) > 3000:
            rng = np.random.default_rng(2021)
            subset_cells = rng.choice(subset_cells, 3000, replace=False)
        local_pdist = gpdist[np.ix_(subset_cells, subset_cells)]
        dists = (np.ravel(local_pdist[np.triu_indices_from(local_pdist, k=1)])
                 .reshape(-1, 1))
        self.local_std = np.std(dists)
        
    def __len__(self):
        return len(self.gene_name) # gene count

    def __getitem__(self, idx):
        gene_name = self.gene_name[idx]
        if self.datastatus=="fit_dataset":
            # random sampling in each epoch
            data_fitting=self.data_fit
            if self.permutation_ratio==1:
                data=data_fitting
            elif (self.permutation_ratio<1) & (self.permutation_ratio>0):
                data=np.random.choice(data_fitting, 
                                      np.round(len(data_fitting)*self.permutation_ratio).astype(int),
                                      replace=False)
            else:
                print('sampling ratio is wrong!')
            
            c = self.c[data, idx]
            u = self.u[data, idx]
            s = self.s[data, idx]
            
            
            embedding = self.embedding[data]
            embedding1 = embedding[:, 0]
            embedding2 = embedding[:, 1]
            
            
        elif self.datastatus=="predict_dataset":
            data=self.data_predict
            c = self.c[data, idx]
            u = self.u[data, idx]
            s = self.s[data, idx]
                        
            embedding = self.embedding[data]
            embedding1 = embedding[:, 0]
            embedding2 = embedding[:, 1]
            
        c = c.astype(np.float32)
        u = u.astype(np.float32)
        s = s.astype(np.float32)
        
        cmax = np.float32(max(c)) 
        umax = np.float32(max(u)) 
        smax = np.float32(max(s))

        if smax <=1e-4:
            return [gene_name]
        elif umax <=1e-4:
            return [gene_name]
        elif cmax <1e-4:
            return [gene_name]
        
        u = u/umax
        s = s/smax
        
        if self.rna_only:
            c = np.ones_like(u)
        else:
            c = c/cmax
        
        # add embedding
        embedding1 = np.array(embedding1.astype(np.float32))
        embedding2 = np.array(embedding2.astype(np.float32))


        return c, u, s, embedding1, embedding2, gene_name, self.local_std



class feedData(pl.LightningDataModule):
    '''
    load training and test data
    '''
    def __init__(self, c, u, s, embedding,
                 rna_only=False, data_fit=None, data_predict=None, gene_name=None,
                 permutation_ratio=1, norm_data=True, norm_cell_distribution=False):
        super().__init__()

        self.fit_dataset = getItem(c, u, s, embedding,
                                   rna_only=rna_only,
                                   data_fit=data_fit, data_predict=data_predict,
                                   gene_name=gene_name,
                                   datastatus="fit_dataset", permutation_ratio=permutation_ratio, norm_data=norm_data, norm_cell_distribution=norm_cell_distribution)
        
        self.predict_dataset = getItem(c, u, s, embedding,
                                       rna_only=rna_only, 
                                       data_fit=data_fit, data_predict=data_predict,
                                       gene_name=gene_name,
                                       datastatus="predict_dataset", permutation_ratio=permutation_ratio,norm_data=norm_data)

    def subset(self, indices):
        import copy
        temp = copy.copy(self)
        temp.fit_dataset = Subset(self.fit_dataset, indices)
        temp.predict_dataset = Subset(self.predict_dataset, indices)
        return temp

    def train_dataloader(self):
        return DataLoader(self.fit_dataset, num_workers=0)
    def val_dataloader(self):
        return DataLoader(self.fit_dataset, num_workers=0)
    def test_dataloader(self):
        return DataLoader(self.predict_dataset, num_workers=0,)


def build_datamodule(c, u, s, embedding,
                     rna_only=False,
                     speed_up=True,
                     norm_data=True,
                     permutation_ratio=None, 
                     norm_cell_distribution=False, 
                     gene_list=None,
                     gene_name=None,
                     downsample_method='neighbors',
                     n_neighbors_downsample=10,
                     step=(200,200),
                     downsample_target_amount=None):

    '''
    get data from target gene
    gene is annotated as index using gene_list
    set fitting data, data to be predicted, and sampling ratio when fitting
    '''
    step_i=step[0]
    step_j=step[1]
    
    if not gene_list is None:
        c = c[:, gene_list]
        u = u[:, gene_list]
        s = s[:, gene_list]    
    unsampled_idx = np.arange(c.shape[0])
     
    if speed_up:
        _, sampling_idx, _ = downsampling_embedding(c, u, s, embedding, 
                            para=downsample_method,
                            target_amount=downsample_target_amount,
                            step=(step_i,step_j),
                            n_neighbors=n_neighbors_downsample,
                            projection_neighbor_choice='embedding')
        
        feed_data = feedData(c, u, s, embedding, 
                             rna_only=rna_only, 
                             data_fit = sampling_idx, data_predict=unsampled_idx,
                             gene_name = gene_name, 
                             permutation_ratio=permutation_ratio, norm_data=norm_data,
                             norm_cell_distribution=norm_cell_distribution) # default 
    else:
        feed_data = feedData(c, u, s, embedding,
                             rna_only=rna_only, 
                             data_fit=unsampled_idx, data_predict=unsampled_idx,
                             gene_name=gene_name, 
                             permutation_ratio=permutation_ratio, norm_data=norm_data,
                             norm_cell_distribution=norm_cell_distribution) 

    return(feed_data)

class MOFlow():
    def __init__(self,
                 adata, 
                 adata_atac=None,
                 rna_only=False,
                 gene_list=None,
                 min_epoches=20,
                 max_epoches=200, 
                 check_val_every_n_epoch=10,
                 patience=3,
                 learning_rate=0.001,
                 n_neighbors=40,
                 embed='X_umap',
                 permutation_ratio=0.125,
                 speed_up=True,
                 norm_data=True,
                 norm_cell_distribution=True,
                 save_path=None,
                 folder_name=None,
                 device=None):
        """Velocity estimation for each cell.
        
        Arguments
        ---------
        adata: `anndata.AnnData`
            anndata that contains the unspliced abundance, spliced abundance, embedding space, and cell type information. 
        adata_atac: `anndata.AnnData`
            anndata that contains the chromatin accessibility.
        gene_list: optional, `list` (default: None)
            Gene list for velocity estimation. `None` if to estimate the velocity of all genes.
        min_epoches: optional, `int` (default: 20)
            Minimum epoches to update the network
        max_epoches: optional, `int` (default: 200)
            Stop to update the network once this number of epochs is reached.
        check_val_every_n_epoch: optional, `int` (default: 10)
            Check loss every n train epochs.
        patience: optional, `int` (default: 3)
            Number of checks with no improvement after which training will be stopped.
        permutation_ratio: optional, `float` (default: 0.125)
            Sampling ratio of cells in each epoch when training each gene.
        speed_up: optional, `bool` (default: True)
            `True` if speed up by downsampling cells. `False` if to use all cells to train the model.
        norm_data: optional, `bool` (default: True)
            `True` if normalize unsplice (and splice) reads by dividing max value of unspliced (and spliced) reads.
        norm_cell_distribution: optional, `bool` (default: True)
            `True` if the bias of cell distribution is to be removed on embedding space (many cells share the same position of unspliced (and spliced) reads).
        n_jobs: optional, `int` (default: -1)
            The maximum number of concurrently running jobs.
        save_path: optional, `str` (default: None)
            Path to save the result of velocity estimation.
        save_path: optional, `str` (default: None)
            folder name to save the result of velocity estimation.
        Returns
        -------
        loss_df: `pandas.DataFrame`
            The record of loss.
        """
        
        # set output dir

        if save_path is None:
            save_path=os.getcwd()
        if folder_name is None:
            folder_name = 'MoFlow_velocity_'

        os.makedirs(os.path.join(save_path,folder_name), exist_ok = True)
        save_path=os.path.join(save_path,folder_name)
        os.makedirs(os.path.join(save_path,'TEMP'), exist_ok = True)
        
        self.save_path = save_path
        if os.path.exists(save_path):
            [os.remove(f) for f in glob.glob(f'{self.save_path}/TEMP/*.csv')]
            
        print('Using '+ save_path +' as the output path.')
        
        # set gene_list if not given
        if gene_list is None:
            gene_list=adata.var_names
        else:
            adata = adata[:, adata.var_names.isin(gene_list)]
            if not rna_only:
                adata_atac = adata_atac[:, adata_atac.var_names.isin(gene_list)]
            gene_list = list(list(set(adata.var_names).intersection(set(gene_list))))
        gene_list_ = np.arange(len(gene_list))
        self.G = len(gene_list)
        self.gene_list = gene_list_ #as index
        self.gene_name = adata.var_names #gene name
        self.gene_list_batch = None
        
        self.rna_only = rna_only
        self.speed_up = speed_up
        self.norm_data = norm_data
        self.permutation_ratio = permutation_ratio
        self.norm_cell_distribution = norm_cell_distribution
        self.min_epoches = min_epoches
        self.max_epoches = max_epoches
        self.check_val_every_n_epoch = check_val_every_n_epoch
        self.patience = patience
        self.n_neighbors = n_neighbors
        self.learning_rate = learning_rate

        if device is None:
            self.accelerator="cpu"
            self.devices="auto"
        else:
            self.accelerator="gpu"
            self.devices="-1"

        u, s = adata.layers['Mu'], adata.layers['Ms'] 
        if rna_only:
            c = np.ones_like(u)
        else:
            c = adata_atac.layers['Mc']

        self.c = c.toarray() if sparse.issparse(c) else c
        self.u = u.toarray() if sparse.issparse(u) else u
        self.s = s.toarray() if sparse.issparse(s) else s
        try:
            self.embedding = adata.obsm[embed]
        except KeyError:
            print("Embedding not found! Set to None.")
            self.embedding = np.nan*np.ones((adata.n_obs, 2))
            
    def _train_thread(self,
                      datamodule, 
                      data_indices,
                      model_save_path=None):
        try:
            seed = 0
            torch.manual_seed(seed)
            random.seed(seed)
            np.random.seed(seed)

            #decide model
            selected_data = datamodule.subset(data_indices)
            
            #c, u, s, embedding1, embedding2, gene_name, local_std = selected_data.predict_dataset.__getitem__(0)
            getitem = selected_data.predict_dataset.__getitem__(0)
            if len(getitem)==1:
                gene_name = getitem[0]
                return gene_name
            else: c, u, s, embedding1, embedding2, gene_name, local_std = getitem
                
            embedding = np.stack((embedding1, embedding2), axis=1)

            # iniate network (DNN_layer) and loss function (DynamicModule)

            model_ = select_initial_net(c, u, s, gene_name, self.rna_only, embedding, local_std)
            
            if model_ == 'low_quality':
                return gene_name
            elif self.rna_only:
                direction_, margin_ = None, None
            else:
                model_, direction_, margin_ = model_

            backbone = DNN_module(DNN_layer(self.rna_only, direction_, self.min_epoches),
                                n_neighbors=self.n_neighbors, rna_only=self.rna_only,
                                )
            model = ltModule(backbone=backbone, learning_rate=self.learning_rate)
            #if self.rna_only:
            #    model.load(model_)


            early_stop_callback = EarlyStopping(monitor="loss", min_delta=0.0, patience=self.patience, mode='min')

            if self.check_val_every_n_epoch is None:
                # not use early stop
                trainer = pl.Trainer(
                    max_epochs=self.max_epoches, 
                    min_epochs = self.min_epoches,
                    reload_dataloaders_every_n_epochs=1, 
                    log_every_n_steps=self.max_epoches+1,
                    logger = False,
                    enable_checkpointing = False,
                    enable_model_summary=False,
                    accelerator=self.accelerator,
                    devices=self.devices,
                    enable_progress_bar=False,
                    )
            else:
                # use early stop
                trainer = pl.Trainer(
                    max_epochs=self.max_epoches, 
                    min_epochs = self.min_epoches,
                    reload_dataloaders_every_n_epochs=1, 
                    log_every_n_steps=self.max_epoches+1,
                    logger = False,
                    enable_checkpointing = False,
                    check_val_every_n_epoch = self.check_val_every_n_epoch,
                    enable_model_summary=False,
                    callbacks=[early_stop_callback],
                    accelerator=self.accelerator,
                    devices=self.devices,
                    enable_progress_bar=False,
                    )

            if self.max_epoches > 0:
                trainer.fit(model, selected_data)   # train network

            trainer.test(model, selected_data, verbose=False)    # predict

            if(model_save_path != None):
                model.save(model_save_path)

            loss = model.validation_loss
            moflow_df = model.test_df


            if(model_save_path != None):
                model.save(model_save_path)

            moflow_df.insert(0, "direction", direction_)
            moflow_df.insert(0, "slope", margin_)
            
            moflow_df.to_csv(os.path.join(self.save_path, 'TEMP', ('moflow_estimation_'+gene_name+'.csv')), index=False)
                
            return None
        except:
            print(f"[ERROR] Gene {gene_name} failed.")
            return gene_name


            
            
    def velocity(self, adata, 
                 model_save_path=None, n_jobs=-1, save_path=None, file_name=None):
        """
        n_jobs: optional, `int` (default: -1)
            The maximum number of concurrently running jobs.
        """
        if save_path is None:
            save_path=os.getcwd()
        #buring
        
        gene_idx_buring = [self.gene_list[0]]
        
        datamodule=build_datamodule(self.c, self.u, self.s, self.embedding,
                                    self.rna_only, self.speed_up,
                                    self.norm_data,
                                    self.permutation_ratio, self.norm_cell_distribution,
                                    gene_list=gene_idx_buring,
                                    gene_name=self.gene_name[0])

        result = Parallel(n_jobs=n_jobs, backend="loky")(
            delayed(self._train_thread)(
                datamodule = datamodule,
                data_indices=[data_index])
            for data_index in range(0,len(gene_idx_buring)))
        
        # clean directory
        #shutil.rmtree(os.path.join(save_path,'TEMP'))
        #os.mkdir(os.path.join(save_path,'TEMP'))
        
        unpredicted_gene_lst = []
        
        id_ranges=list()
        if n_jobs==-1:
            interval=os.cpu_count()
        else:
            interval=n_jobs
        for i in range(0, self.G, interval):
            idx_start=i
            if self.G < i+interval:
                idx_end = self.G
            else:
                idx_end=i+interval
            id_ranges.append((idx_start,idx_end))


        print('Arranging genes for parallel job.')
        if len(id_ranges)==1:
            if id_ranges==1:
                print(self.G,' gene was arranged to ',len(id_ranges),' portion.')
            else:
                print(self.G,' genes were arranged to ',len(id_ranges),' portion.')
        else: 
            print(self.G,' genes were arranged to ',len(id_ranges),' portions.')

        unpredicted_gene_lst=list()

        for id_range in tqdm(id_ranges, desc="Velocity Estimation", total=len(id_ranges), position=1, leave=False, bar_format='{l_bar}{bar:10}{r_bar}{bar:-10b}'):

            self.gene_list_batch=self.gene_list[id_range[0]:id_range[1]]
            gene_name_batch=self.gene_name[id_range[0]:id_range[1]]
            
            datamodule=build_datamodule(self.c, self.u, self.s, self.embedding,
                                        self.rna_only, self.speed_up, 
                                        self.norm_data, 
                                        self.permutation_ratio, self.norm_cell_distribution,
                                        gene_list=self.gene_list_batch, gene_name=gene_name_batch)

            results = Parallel(n_jobs=n_jobs, backend="loky")(
                delayed(self._train_thread)(
                datamodule = datamodule,
                data_indices=[data_index])
                for data_index in range(0,len(self.gene_list_batch)))
            gene_name_lst=[x for x in result if x is not None]
            
            for i in gene_name_lst:
                unpredicted_gene_lst.append(i)
        if len(unpredicted_gene_lst)!=0:
            self.gene_name = set(self.gene_name) - set(unpredicted_gene_lst)


        
                
        def combine_csv(save_path, files):
            with open(save_path,"wb") as fout:
                # first file:
                with open(files[0], "rb") as f:
                    fout.write(f.read())
                # the rest:    
                for filepath in files[1:]:
                    with open(filepath, "rb") as f:
                        next(f)
                        fout.write(f.read())
            return(pd.read_csv(save_path))
        
        moflow_df = os.path.join(self.save_path,'TEMP', "moflow_estimation*.csv")
        moflow_df_files = glob.glob(moflow_df)
        moflow_df=combine_csv(os.path.join(self.save_path, "moflow_estimation.csv"), moflow_df_files)
        
        moflow_df = moflow_df[moflow_df['gene_name'].isin(self.gene_name)]
        moflow_df = moflow_df.dropna()
        
        c_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='c')
        u_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='u')
        s_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='s')
        vc_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='vc')
        vu_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='vu')
        vs_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='vs')
        alpha_c_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='alpha_c')
        alpha_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='alpha')
        beta_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='beta')
        gamma_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='gamma')
        stateon_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='stateon')
        stateoff_ = moflow_df.pivot_table(index='cellIndex', columns='gene_name', values='stateoff')
        direction_ = moflow_df[['gene_name', 'direction']].drop_duplicates().set_index('gene_name')
        slope_ = moflow_df[['gene_name', 'slope']].drop_duplicates().set_index('gene_name')
        loss = moflow_df[['gene_name', 'loss']].drop_duplicates().set_index('gene_name')
        epoch = moflow_df[['gene_name', 'epoch']].drop_duplicates().set_index('gene_name')
        
   
        gene_name_ = c_.columns.values
        adata = adata[:, adata.var_names.isin(gene_name_)]
        
        adata.layers['Mc'] = c_
        adata.layers['c'] = c_
        adata.layers['u'] = u_
        adata.layers['s'] = s_
        
        adata.layers['alpha'] = alpha_
        adata.layers['beta'] = beta_
        adata.layers['gamma'] = gamma_
        
        adata.layers['alpha_c'] = alpha_c_

        adata.layers['velo_c'] = vc_
        adata.layers['velo_u'] = vu_
        adata.layers['velo_s'] = vs_
        adata.layers['stateon'] = stateon_
        adata.layers['stateoff'] = stateoff_

        adata.var['loss'] = loss
        adata.var['epoch'] = epoch
        adata.var['direction'] = direction_
        adata.var['slope'] = slope_

        vkey='velo_s'
        adata.layers[vkey+'_norm'] = adata.layers[vkey] / np.sum(
            np.abs(adata.layers[vkey]), 0)
        adata.layers[vkey+'_norm'] /= np.mean(adata.layers[vkey+'_norm'])
        
        if file_name is not None:
            print('write data')
            adata.write(f"{save_path}/{file_name}")   
            print('Result is perfectly saved at ' + save_path + '/' + file_name)
        else:
            return adata

            





def select_initial_net(
    c, u, s, gene_name, rna_only, embedding, local_std, fit_slope=True):
    '''
    check if right top conner has cells
    circle.pt is the model for single kinetic
    branch.pt is multiple kinetic
    '''
    s_max=np.max(s)
    u_max = np.max(u)
    c_max = np.max(c)
    s_max_90per = 0.9*s_max
    u_max_90per = 0.9*u_max
    
    if s_max <=1e-4:
        return 'low_quality'
    elif u_max <=1e-4:
        return 'low_quality'
    elif c_max <1e-4:
        return 'low_quality'

    cells_corner = (s>s_max_90per) & (u>u_max_90per)
    if np.sum(cells_corner) > 0.001*u.shape[0]: 
        model_r= 'circle'
    else:
        model_r = 'branch'
    

    if rna_only:
        return model_r
    else:
        
        w_non_zero = ((c >= 0.1 * np.max(c)) &
                      (u >= 0.1 * np.max(u)) &
                      (s >= 0.1 * np.max(s)))
        u_non_zero = u[w_non_zero]
        s_non_zero = s[w_non_zero]
        if len(u_non_zero) < 10:
            low_quality = True
            return 'low_quality'

        # GMM
        w_low = ((np.percentile(s_non_zero, 30) <= s_non_zero) &
                 (s_non_zero <= np.percentile(s_non_zero, 40)))
        if np.sum(w_low) < 10:
            fit_gmm = False
            partial = True
        else:

            pdist = pairwise_distances(
                embedding[w_non_zero, :][w_low, :])
            dists = (np.ravel(pdist[np.triu_indices_from(pdist, k=1)])
                     .reshape(-1, 1))
            model = GaussianMixture(n_components=2, covariance_type='tied',
                                    random_state=2021).fit(dists)
            mean_diff = np.abs(model.means_[1][0] - model.means_[0][0])
            criterion1 = mean_diff > local_std 
            criterion2 = np.all(model.weights_[1] > 0.2)

            if criterion1 and criterion2:
                partial = False
            else:
                partial = True

        # steady-state slope
        wu = u >= np.percentile(u_non_zero, 95)
        ws = s >= np.percentile(s_non_zero, 95)
        ss_u = u[wu | ws]
        ss_s = s[wu | ws]
        if np.all(ss_u == 0) or np.all(ss_s == 0):
            return 'low_quality'
        gamma = np.dot(ss_u, ss_s) / np.dot(ss_s, ss_s)

        # thickness of phase portrait
        u_norm = u_non_zero / np.max(u)
        s_norm = s_non_zero / np.max(s)
        exp = np.hstack((np.reshape(u_norm, (-1, 1)),
                         np.reshape(s_norm, (-1, 1))))
        U, S, Vh = np.linalg.svd(exp)
        thickness = S[1]

        # slope-based direction decision
        with np.errstate(divide='ignore', invalid='ignore'):
            slope = u /s
        non_naninf = (~np.isnan(slope))&(~np.isinf(slope))
        slope = slope[non_naninf]
        on = slope >= gamma
        off = slope < gamma
        if len(ss_u) < 10 or len(u_non_zero) < 10:
            fit_slope = False
            direction = 'complete'
        
        else:
            slope_ = u_non_zero / s_non_zero
            on_ = slope_ >= gamma
            off_ = slope_ < gamma
            on_dist = np.sum((u_non_zero[on_] - gamma * s_non_zero[on_])**2)
            off_dist = np.sum((gamma * s_non_zero[off_] - u_non_zero[off_])**2)

            if thickness < 1.5:
                narrow = True
            else:
                narrow = False

            if on_dist > 10 * off_dist:
                direction = 'on'
                partial = True
            elif off_dist > 10  * on_dist:
                direction = 'off'
                partial = True
            else:
                if partial is True:
                    if on_dist > 3 * off_dist:
                        direction = 'on'
                    elif off_dist > 3 * on_dist:
                        direction = 'off'
                    else:
                        if narrow:
                            direction = 'on'
                        else:
                            direction = 'complete'
                            partial = False
                else:
                    if narrow:
                        direction = ('off'
                                          if off_dist > 2  * on_dist
                                          else 'on')
                        partial = True
                    else:
                        direction = 'complete'

        # model pre-determination
        if direction == 'on':
            model_ = 1
        elif direction == 'off':
            model_ = 2
        else: #complete
            c_high = c >= np.mean(c) + 2 * np.std(c)
            c_high = c_high[non_naninf]
            if np.sum(c_high) < 10:
                c_high = c >= np.mean(c) + np.std(c)
                c_high = c_high[non_naninf]
            if np.sum(c_high) < 10:
                c_high = c >= np.percentile(c, 90)
                c_high = c_high[non_naninf]
            if np.sum(c[non_naninf][c_high] == 0) > 0.5*np.sum(c_high):
                return 'low_quality'
            c_high_on = np.sum(c_high & on)
            c_high_off = np.sum(c_high & off)
            if c_high_on > c_high_off:
                model_ = 1
            else:
                model_ = 2
        
    
    
    return(model_r, direction, np.median(slope))





from .moflow import MOFlow
from .plot_gene import scatter_gene
from .plot_gene import dynamic_plot
from .plot_cell import scatter_cell
from .eval import cross_boundary_correctness
from .eval import inner_cluster_coh
from .eval_dtw import plot_dtw
from .simulation import simulate

__all__ = [
]






import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from scipy import sparse
from sklearn.neighbors import NearestNeighbors
import seaborn as sns




def sampling_neighbors(plotdata,step=(30,30),percentile=25):

    from scipy.stats import norm
    def gaussian_kernel(X, mu = 0, sigma=1):
        return np.exp(-(X - mu)**2 / (2*sigma**2)) / np.sqrt(2*np.pi*sigma**2)
    embed_len = plotdata.shape[1]
    
    if len(step) != embed_len:
        step = np.repeat(step[0], embed_len)
    grs = []
    for dim_i in range(plotdata.shape[1]):
        m, M = np.min(plotdata[:, dim_i]), np.max(plotdata[:, dim_i])
        m = m - 0.025 * np.abs(M - m)
        M = M + 0.025 * np.abs(M - m)
        gr = np.linspace(m, M, step[dim_i])
        grs.append(gr)
    meshes_tuple = np.meshgrid(*grs)
   
    gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T
    gridpoints_coordinates = gridpoints_coordinates + norm.rvs(loc=0, scale=0.15, size=gridpoints_coordinates.shape)
    
    np.random.seed(10) # set random seed
    
    nn = NearestNeighbors()

    neighbors_1 = min((plotdata.shape[0]-1), 20)
    nn.fit(plotdata)
    dist, ixs = nn.kneighbors(gridpoints_coordinates, neighbors_1)

    ix_choice = ixs[:,0].flat[:]
    ix_choice = np.unique(ix_choice)

    nn = NearestNeighbors()

    neighbors_2 = min((plotdata.shape[0]-1), 20)
    nn.fit(plotdata)
    dist, ixs = nn.kneighbors(plotdata[ix_choice, :embed_len], neighbors_2)
    
    density_extimate = gaussian_kernel(dist, mu=0, sigma=0.5).sum(1)
    bool_density = density_extimate > np.percentile(density_extimate, percentile)
    ix_choice = ix_choice[bool_density]
    return(ix_choice)

    
def scatter_gene(adata,
                 gene_list,
                 by='us',
                 color_by='celltype',
                 colors=None,
                 n_cols=5,
                 axis_on=True,
                 frame_on=True,
                 velocity_arrows=False,
                 downsample=1,
                 percentile=0,
                 figsize=None,
                 pointsize=2,
                 alpha=.5,
                 cmap='coolwarm',
                 view_3d_elev=None,
                 view_3d_azim=None,
                 full_name=False,
                 log_scale=False,
                 **kwargs

                 ):
    """Gene scatter plot.

    This function plots phase portraits of the specified plane.

    Parameters
    ----------
    data: :class:`~anndata.AnnData` 
        Data result from relay velocity model.
    gene_list: `str`,  list of `str`
        List of gene_list to plot.
    by: `str` (default: `us`)
        Plot unspliced-spliced plane if `us`. Plot chromatin-unspliced plane
        if `cu`.
        Plot 3D phase portraits if `cus`.
    color_by: `str` (default: `celltype`)
        Color by the four potential states if `celltype`. Other common values are
        leiden, louvain, cluster, etc.
        If not `celltype`, the color field must be present in `.uns`, which can be
        pre-computed with `scanpy.pl.scatter`.
        Color by can also be common values like `c` in `.layers`, which displays the log
        accessibility on U-S phase portraits.
    n_cols: `int` (default: 5)
        Number of columns to plot on each row.
    axis_on: `bool` (default: `True`)
        Whether to show axis labels.
    frame_on: `bool` (default: `True`)
        Whether to show plot frames.
    velocity_arrows: `bool` (default: `False`)
        Whether to show velocity arrows of cells on the phase portraits.
    downsample: `int` (default: 1)
        How much to downsample the cells. The remaining number will be
        `1/downsample` of original.
    figsize: `tuple` (default: `None`)
        Total figure size.
    pointsize: `float` (default: 2)
        Point size for scatter plots.
    linewidth: `float` (default: 2)
        Line width for connected anchors.
    cmap: `str` (default: `coolwarm`)
        Color map for log accessibilities or other continuous color keys when
        plotting on U-S plane.
    view_3d_elev: `float` (default: `None`)
        Matplotlib 3D plot `elev` argument. `elev=90` is the same as U-S plane,
        and `elev=0` is the same as C-U plane.
    view_3d_azim: `float` (default: `None`)
        Matplotlib 3D plot `azim` argument. `azim=270` is the same as U-S
        plane, and `azim=0` is the same as C-U plane.
    full_name: `bool` (default: `False`)
        Show full names for chromatin, unspliced, and spliced rather than
        using abbreviated terms c, u, and s.
    """
    from pandas.api.types import is_numeric_dtype, is_categorical_dtype


    if by not in ['us', 'cu', 'cus']:
        raise ValueError("'by' argument must be one of ['us', 'cu', 'cus']")

    #set colorcode
    if color_by in adata.layers and is_numeric_dtype(adata.layers[color_by]):
        types = None
        colors = None
    elif color_by in adata.obs and is_numeric_dtype(adata.obs[color_by]):
        types = None
        colors = adata.obs[color_by].values
    elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]):
        if colors is not None:
            types = adata.obs[color_by].cat.categories
            if set(colors.keys()) & set(types) != set(types):
                raise ValueError("Currently, colors keys don't contain all categories of types.")

        elif color_by+'_colors' in adata.uns.keys():
            types = adata.obs[color_by].cat.categories
            colors = adata.uns[f'{color_by}_colors']
        else:
            raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata'
                         ' obs, and the colors of categories can be found in'
                         ' adata uns.')
    elif color_by == 'state' and 'fit_state' not in adata.layers.keys():
        raise ValueError('fit_state is not found. Please run '
                         'recover_dynamics_chrom function first or provide a '
                         'valid color key.')    
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata'
                         ' obs, and the colors of categories can be found in'
                         ' adata uns.')
     
    
    downsample = np.clip(int(downsample), 1, 10)
    gene_list = np.array(gene_list)
    missing_genes = gene_list[~np.isin(gene_list, adata.var_names)]
    
    #if len(missing_genes) > 0: 
    #    logg.update(f'{missing_genes} not found', v=0)
    gene_list = gene_list[np.isin(gene_list, adata.var_names)]
    gn = len(gene_list)

    #Set figure and axes
    if gn == 0:
        return
    if gn < n_cols:
        n_cols = gn
    if by == 'cus':
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(4*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize,
                                subplot_kw={'projection': '3d'})
    else:
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3.2*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize)
    fig.patch.set_facecolor('white')
                     
    
    for g, gene in enumerate(gene_list):
        c = adata[:, gene].layers['Mc'].copy()
        u = adata[:, gene].layers['Mu'].copy()
        s = adata[:, gene].layers['Ms'].copy()

        c = c.A if sparse.issparse(c) else c
        u = u.A if sparse.issparse(u) else u
        s = s.A if sparse.issparse(s) else s
        c, u, s = np.ravel(c), np.ravel(u), np.ravel(s)

        if percentile>0:
            if by=='us':
                ix_choice = sampling_neighbors(np.array([u, s]).T, percentile=percentile)
            elif by=='cu':
                ix_choice = sampling_neighbors(np.array([c, u]).T, percentile=percentile)
            else:
                ix_choice = sampling_neighbors(np.array([c, u, s]).T, percentile=percentile)
        else:
            ix_choice = np.arange(c.shape[0])
                
        
        
        if velocity_arrows:
            if 'velo_u' in adata.layers.keys():
                vu = adata[:, gene].layers['velo_u'].copy()
            else:
                vu = np.zeros(adata.n_obs)
            max_u = np.max([np.max(u), 1e-6])
            u /= max_u
            vu = np.ravel(vu)
            vu /= np.max([np.max(np.abs(vu),), 1e-6])
            
            if 'velo_s' in adata.layers.keys():
                vs = adata[:, gene].layers['velo_s'].copy()
            else:
                vs = adata[:, gene].layers['velocity'].copy()
            max_s = np.max([np.max(s), 1e-6])
            s /= max_s
            vs = np.ravel(vs)
            vs /= np.max([np.max(np.abs(vs)), 1e-6])
            if 'velo_c' in adata.layers.keys():
                vc = adata[:, gene].layers['velo_c'].copy()
                max_c = np.max([np.max(c), 1e-6])
                c /= max_c
                vc = np.ravel(vc)
                vc /= np.max([np.max(np.abs(vc)), 1e-6])
            
            
        row = g // n_cols
        col = g % n_cols
        ax = axs[row, col]
        
        
        
        if types is not None:
            for i in range(len(types)):
                filt = adata.obs[color_by] == types[i]
                filt = np.arange(adata.n_obs)[np.ravel(filt)]

                if type(colors) == dict:
                    color = colors[types[i]]
                else:
                    color = colors[i]
                
                if by == 'us':
                    
                    if velocity_arrows:
                        
                        filt = list(set(filt) & set(ix_choice))
                        ax.quiver(s[filt], u[filt], vs[filt], vu[filt], color=color, alpha=alpha,
			scale_units='xy', scale=10, width=0.005, headwidth=4, headaxislength=5.5, **kwargs)
                    else:
                        ax.scatter(s[filt][::downsample], u[filt][::downsample], s=pointsize,
                                   c=color, alpha=alpha, **kwargs)
                        

                elif by == 'cu':
                    if velocity_arrows:
                        filt = list(set(filt) & set(ix_choice))
                        ax.quiver(u[filt], c[filt], vu[filt], vc[filt], color=color, alpha=alpha,
			scale_units='xy', scale=10, width=0.005, headwidth=4, headaxislength=5.5, **kwargs)
                    else:
                        ax.scatter(u[filt][::downsample], c[filt][::downsample], s=pointsize,
                                   c=color, alpha=alpha, **kwargs)
      
                else:
                    
                    if velocity_arrows:
                        filt = list(set(filt) & set(ix_choice))
                        ax.quiver(s[filt], u[filt],c[filt], 
                                  vs[filt], vu[filt], vc[filt],
                                  color=color, alpha=0.4, length=0.1,
                                arrow_length_ratio=0.5, normalize=True, **kwargs)
                    else:
                        ax.scatter(s[filt][::downsample], u[filt][::downsample], c[filt][::downsample], s=pointsize,c=color, alpha=alpha, **kwargs)
                    
                        

        elif colors is None:
            color = adata[:, gene].layers[color_by].copy()
            color = color.A if sparse.issparse(color) else color
            color = color.ravel()
            
            outlier = 99.8
            non_zero = (u > 0) & (s > 0) & (c > 0)
            non_outlier = u < np.percentile(u, outlier)
            non_outlier &= s < np.percentile(s, outlier)
            non_outlier &= c < np.percentile(c, outlier)
            color -= np.min(color)
            color /= np.max(color)
            
            if by == 'us':
                if velocity_arrows:
                    ax.quiver(s[non_zero & non_outlier][::downsample], 
                            u[non_zero & non_outlier][::downsample],
                            vs[non_zero & non_outlier][::downsample],
                            vu[non_zero & non_outlier][::downsample],
                            
                            np.log1p(color[non_zero & non_outlier][::downsample]), cmap=cmap, alpha=0.7, 
                scale_units='xy', scale=10, width=0.005, headwidth=4, headaxislength=5.5, **kwargs)
                else:
                    ax.scatter(s[non_zero & non_outlier][::downsample], u[non_zero & non_outlier][::downsample], s=pointsize,
                               c=color[non_zero & non_outlier][::downsample],
                        alpha=alpha, cmap=cmap, **kwargs)
            elif by =='cu':
                if velocity_arrows:
                    ax.quiver(u[non_zero & non_outlier][::downsample], 
                            c[non_zero & non_outlier][::downsample],
                            vu[non_zero & non_outlier][::downsample],
                            vc[non_zero & non_outlier][::downsample],
                            color[non_zero & non_outlier][::downsample],
                            cmap=cmap, alpha=0.7, 
                scale_units='xy', scale=10, width=0.005, headwidth=4, headaxislength=5.5, **kwargs)
                else:
                    ax.scatter(u[non_zero & non_outlier][::downsample], c[non_zero & non_outlier][::downsample], s=pointsize, c=color[non_zero & non_outlier][::downsample],
                        alpha=alpha, cmap=cmap, **kwargs)
                    
            else:
                if velocity_arrows:
                    ax.quiver(s[non_zero & non_outlier][::downsample], u[non_zero & non_outlier][::downsample],c[non_zero & non_outlier][::downsample], 
                                vs[non_zero & non_outlier][::downsample], vu[non_zero & non_outlier][::downsample], vc[non_zero & non_outlier][::downsample],
                                c=np.log1p(color[non_zero & non_outlier][::downsample]),
                                                alpha=0.4, length=0.1,
                            arrow_length_ratio=0.5, normalize=True, **kwargs)
                else:
                    ax.scatter(s[non_zero & non_outlier][::downsample], u[non_zero & non_outlier][::downsample], c[non_zero & non_outlier][::downsample], s=pointsize,
                               c=np.log1p(color[non_zero & non_outlier][::downsample]), alpha=alpha, **kwargs)
                
                    
        else:
            if by == 'us':
                if velocity_arrows:
                    ax.quiver(s[::downsample], u[::downsample],
                                  vs[::downsample],
                                  vu[::downsample],
                                colors[::downsample], alpha=0.5,
                                scale_units='xy', scale=10, width=0.005,
                                headwidth=4, headaxislength=5.5, cmap=cmap, **kwargs)
                else:
                    ax.scatter(s[::downsample], u[::downsample], s=pointsize,
                                c=colors[::downsample], alpha=0.7, cmap=cmap, **kwargs)

            elif by == 'cu':
                if velocity_arrows:
                    ax.quiver(u[::downsample],
                                c[::downsample],
                                vu[::downsample],
                                vc[::downsample],
                                
                                colors[::downsample], alpha=0.5,
                                scale_units='xy', scale=10, width=0.005,
                                headwidth=4, headaxislength=5.5, cmap=cmap, **kwargs)
                else:
                    ax.scatter(u[::downsample], c[::downsample], s=pointsize,
                                c=colors[::downsample], alpha=0.7, cmap=cmap, **kwargs)

                    
            else:
                ax.scatter(s[::downsample], u[::downsample],
                                c[::downsample], s=pointsize,
                                c=colors[::downsample], alpha=0.7, cmap=cmap, **kwargs)
                if velocity_arrows:
                    ax.quiver(s[::downsample],
                                u[::downsample], c[::downsample],
                                vs[::downsample],
                                vu[::downsample],
                                vc[::downsample],
                                colors[::downsample], alpha=0.4, length=0.1,
                                arrow_length_ratio=0.5, normalize=True,
                                cmap=cmap, **kwargs)


        if by == 'cus' and \
                (view_3d_elev is not None or view_3d_azim is not None):
            ax.view_init(elev=view_3d_elev, azim=view_3d_azim)
        title = gene
        ax.set_title(f'{title}', fontsize=11)
        
        if by == 'us':
            ax.set_xlabel('spliced' if full_name else 's')
            ax.set_ylabel('unspliced' if full_name else 'u')
        elif by == 'cu':
            ax.set_xlabel('unspliced' if full_name else 'u')
            ax.set_ylabel('chromatin' if full_name else 'c')
        elif by == 'cus':
            ax.set_xlabel('spliced' if full_name else 's')
            ax.set_ylabel('unspliced' if full_name else 'u')
            ax.set_zlabel('chromatin' if full_name else 'c')
        if by in ['us', 'cu']:
            if not axis_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
            if not frame_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.set_frame_on(False)
            if log_scale:
                ax.set_xscale('log')
                ax.set_yscale('log')
        elif by == 'cus':
            if not axis_on:
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.set_zlabel('')
                ax.xaxis.set_ticklabels([])
                ax.yaxis.set_ticklabels([])
                ax.zaxis.set_ticklabels([])
            if not frame_on:
                ax.xaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.yaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.zaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.xaxis._axinfo['tick']['inward_factor'] = 0
                ax.xaxis._axinfo['tick']['outward_factor'] = 0
                ax.yaxis._axinfo['tick']['inward_factor'] = 0
                ax.yaxis._axinfo['tick']['outward_factor'] = 0
                ax.zaxis._axinfo['tick']['inward_factor'] = 0
                ax.zaxis._axinfo['tick']['outward_factor'] = 0
            if log_scale:
                ax.set_xscale('log')
                ax.set_yscale('log')
                
       
    for i in range(col+1, n_cols):
        fig.delaxes(axs[row, i])
    fig.tight_layout()
    return fig, axs

def dynamic_plot(adata,
                 gene_list,
                 by='expression',
                 time_key= 'velo_s_pseudotime',
                 color_by='celltype',
                 gene_time=False,
                 axis_on=True,
                 frame_on=True,
                 downsample=1,
                 figsize=None,
                 pointsize=2,
                 cmap='coolwarm',
                 **kwargs
                 ):
    """Gene dynamics plot.

    This function plots accessibility, expression, or velocity by time.

    Parameters
    ----------
    adata: :class:`~anndata.AnnData` or `~mudata.MuData`
        Data result from dynamics recovery.
    gene_list: `str`,  list of `str`
        List of gene_list to plot.
    by: `str` (default: `expression`)
        Plot accessibilities and expressions if `expression`. Plot velocities
        if `velocity`.
    color_by: `str` (default: `celltype`)
        Color by the four potential states if `state`. Other common values are
        leiden, louvain, cluster, etc.
        If not `celltype`, the color field must be present in `.uns`, which can
        be pre-computed with `scanpy.pl.scatter`.
    time_key: `str` (default: `velo_s_pseudotime`)
        Time key to plot.
        If gene_time is True, the time field must be present in `.layers`. 
        Unless, it must be present in `.obs`. 
    gene_time: `bool` (default: `False`)
        Whether to use individual gene fitted time, or shared global latent
        time.
    axis_on: `bool` (default: `True`)
        Whether to show axis labels.
    frame_on: `bool` (default: `True`)
        Whether to show plot frames.
    downsample: `int` (default: 1)
        How much to downsample the cells. The remaining number will be
        `1/downsample` of original.
    full_range: `bool` (default: `False`)
        Whether to show the full time range of velocities before smoothing or
        subset to only smoothed range.
    figsize: `tuple` (default: `None`)
        Total figure size.
    pointsize: `float` (default: 2)
        Point size for scatter plots.
    cmap: `str` (default: `coolwarm`)
        Color map for continuous color key.
    """
    
    from pandas.api.types import is_numeric_dtype, is_categorical_dtype
    

    if by not in ['expression', 'velocity', 'expectation']:
        raise ValueError('"by" must be either "expression", "expectation" or "velocity".')
    elif color_by in adata.obs and is_numeric_dtype(adata.obs[color_by]):
        types = None
        colors = adata.obs[color_by].values
    elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]) \
            and color_by+'_colors' in adata.uns.keys():
        types = adata.obs[color_by].cat.categories
        colors = adata.uns[f'{color_by}_colors']
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata '
                         'obs, and the colors of categories can be found in '
                         'adata uns.')

    downsample = np.clip(int(downsample), 1, 10)
    gene_list = np.array(gene_list)
    missing_genes = gene_list[~np.isin(gene_list, adata.var_names)]

    gene_list = gene_list[np.isin(gene_list, adata.var_names)]
    gn = len(gene_list)
    if gn == 0:
        return
    
    if time_key is None:
        if gene_time:
            time_key = 'fit_t'
        else:
            time_key = 'velo_s_pseudotime'

    

    fig, axs = plt.subplots(gn, 3, squeeze=False, figsize=(10, 2.3*gn)
                            if figsize is None else figsize)
    fig.patch.set_facecolor('white')
    for row, gene in enumerate(gene_list):
        if by == 'expression':
            c = adata[:, gene].layers['c']
            u = adata[:, gene].layers['u']
            s = adata[:, gene].layers['s'] 
        elif by == 'velocity':
            u = adata[:, gene].layers['velo_u']
            s = adata[:, gene].layers['velo_s']
            c = adata[:, gene].layers['velo_c']
        else:
            u = adata[:, gene].layers['uhat']
            s = adata[:, gene].layers['shat']
            c = adata[:, gene].layers['chat']
        
        c = c.A if sparse.issparse(c) else c
        u = u.A if sparse.issparse(u) else u
        s = s.A if sparse.issparse(s) else s
        c, u, s = np.ravel(c), np.ravel(u), np.ravel(s)


        if types is not None:
            time = np.array(adata[:, gene].layers[time_key] if gene_time 
                            else adata.obs[time_key])
            for i in range(len(types)):
                filt = adata.obs[color_by] == types[i]
                filt = np.ravel(filt)
                if np.sum(filt) > 0:
                    axs[row, 0].scatter(time[filt][::downsample],
                                        c[filt][::downsample], s=pointsize,
                                        c=colors[i], alpha=0.6, **kwargs)
                    axs[row, 1].scatter(time[filt][::downsample],
                                        u[filt][::downsample],
                                        s=pointsize, c=colors[i], alpha=0.6, **kwargs)
                    axs[row, 2].scatter(time[filt][::downsample],
                                        s[filt][::downsample], s=pointsize,
                                        c=colors[i], alpha=0.6, **kwargs)
        else:
            time = np.array(adata[:, gene].layers[time_key] if gene_time 
                            else adata.obs[time_key])
            axs[row, 0].scatter(time[::downsample], c[::downsample],
                                s=pointsize,
                                c=colors[::downsample],
                                alpha=0.6, cmap=cmap, **kwargs)
            axs[row, 1].scatter(time[::downsample], u[::downsample],
                                s=pointsize,
                                c=colors[::downsample],
                                alpha=0.6, cmap=cmap, **kwargs)
            axs[row, 2].scatter(time[::downsample], s[::downsample],
                                s=pointsize,
                                c=colors[::downsample],
                                alpha=0.6, cmap=cmap, **kwargs)

        

        

        axs[row, 0].set_title(f'{gene} ATAC' + ('' if by == 'expression'
                              else ' velocity' if by=='velocity' else ' expectation'))
        axs[row, 0].set_xlabel('~t' if by == 'velocity' else 't')
        axs[row, 0].set_ylabel('dc/dt' if by == 'velocity' else 'c')
        axs[row, 1].set_title(f'{gene} unspliced' + ('' if by == 'expression'
                              else ' velocity' if by=='velocity' else ' expectation'))
        axs[row, 1].set_xlabel('~t' if by == 'velocity' else 't')
        axs[row, 1].set_ylabel('du/dt' if by == 'velocity' else 'u')
        axs[row, 2].set_title(f'{gene} spliced' + ('' if by == 'expression'
                              else ' velocity' if by=='velocity' else ' expectation'))
        axs[row, 2].set_xlabel('~t' if by == 'velocity' else 't')
        axs[row, 2].set_ylabel('ds/dt' if by == 'velocity' else 's')

        for j in range(3):
            ax = axs[row, j]
            if not axis_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
            if not frame_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.set_frame_on(False)
        
        
            
    fig.tight_layout()
    return(fig, axs)

def violinplot(adata,
                 gene_list,
                 by='alpha',
                 color_by='celltype',
                 colors=None,
                 axis_on=True,
                 frame_on=True,
                 n_cols=5,
                 figsize=None,
                 type='violin',
                 **kwargs):
    """
    Create violin plots (or box plots) for a list of genes, grouped by a categorical variable (e.g., cell type).

    This function generates individual plots for each gene in the given `gene_list`, showing the distribution of 
    a specific layer (e.g., spliced, unspliced) across different groups (e.g., cell types). The plots can be customized 
    as violin plots or box plots. The function allows flexibility in color schemes, axis visibility, and layout.

    Parameters
    ----------
    adata : `anndata.AnnData`
        Annotated data object containing the gene expression data. The gene expression data should be in the layers 
        of the object (e.g., 'alpha', 'Ms', etc.), and cell labels should be available in `adata.obs`.

    gene_list : `list of str`
        A list of gene names to generate plots for. The function will only plot the genes that are present in 
        `adata.var_names`.

    by : `str`, optional (default: 'alpha')
        The key in the `adata.layers` where the gene expression data is stored (e.g., 'alpha', 'Ms', 'Mu'). 

    color_by : `str`, optional (default: 'celltype')
        The key in `adata.obs` for grouping the cells (e.g., 'celltype', 'cluster'). This is used for coloring the 
        plot based on the categorical variable.

    colors : `dict` or `None`, optional (default: None)
        A dictionary mapping categories to specific colors. If not provided, the function will attempt to use the 
        colors stored in `adata.uns` for the `color_by` variable.

    axis_on : `bool`, optional (default: True)
        If `False`, the axis labels and ticks will be hidden. If `True`, they will be displayed.

    frame_on : `bool`, optional (default: True)
        If `False`, the box/violin frame will not be shown.

    n_cols : `int`, optional (default: 5)
        The number of columns to arrange the plots in. If the number of genes is smaller than `n_cols`, the number of 
        columns will be adjusted automatically.

    figsize : `tuple`, optional (default: None)
        The size of the figure. If not provided, the figure size will be automatically adjusted based on the number of 
        columns and rows.

    type : `str`, optional (default: 'violin')
        The type of plot to generate. Can be either 'violin' or 'box'. 'violin' will produce violin plots, while 'box' 
        will produce box plots.

    **kwargs : keyword arguments
        Additional arguments passed to `sns.violinplot()` or `sns.boxplot()`. These can be used for further customization 
        of the plots.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The figure object containing the generated plots.

    axs : `matplotlib.axes.Axes`
        The axes of the subplots.

    Notes
    -----
    - The number of rows and columns in the subplot grid is adjusted based on the number of genes in `gene_list`.
    - The plots are grouped by the `color_by` variable, and each group is assigned a different color, as specified by 
      the `colors` parameter or the color scheme available in `adata.uns`.

    Examples
    --------
    # Example usage:
    violinplot(adata, gene_list=['GeneA', 'GeneB'], by='alpha', color_by='celltype', n_cols=4)
    """
    
    gene_list = np.array(gene_list)
    missing_genes = gene_list[~np.isin(gene_list, adata.var_names)]
    
    from pandas.api.types import is_categorical_dtype

    #set colorcode
    if color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]):
        if colors is not None:
            types = adata.obs[color_by].cat.categories
            if set(colors.keys()) & set(types) != set(types):
                raise ValueError("Currently, colors keys don't contain all categories of types.")

        elif color_by+'_colors' in adata.uns.keys():
            types = adata.obs[color_by].cat.categories
            colors = adata.uns[f'{color_by}_colors']
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'categorical available in adata obs. ')

    gene_list = gene_list[np.isin(gene_list, adata.var_names)]
    gn = len(gene_list)
    if gn == 0:
        return
    elif gn < n_cols:
        n_cols = gn

    fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3.2*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize)
    fig.patch.set_facecolor('white')
                     
    
    for g, gene in enumerate(gene_list):
        row = g // n_cols
        col = g % n_cols
        ax = axs[row, col]
        df = pd.DataFrame(adata[:, gene].layers[by], 
                          columns=[by],
                          index=adata.obs_names)
        df['cluster'] = adata.obs[color_by]
        if type == 'violin':
            if colors is None:
                sns.violinplot(data=df, x='cluster', y=by,
                           palette=colors, ax=ax, scale='width', **kwargs)
            else:
                sns.violinplot(data=df, x='cluster', y=by,
                           palette=colors, ax=ax, scale='width',**kwargs)
        elif type == 'box':
            if colors is None:
                sns.boxplot(data=df, x='cluster', y=by,
                           palette=colors, ax=ax, **kwargs)
            else:
                sns.boxplot(data=df, x='cluster', y=by,
                           palette=colors, ax=ax, **kwargs)
        ax.set_title(gene)
        if not axis_on:
            ax.xaxis.set_ticks_position('none')
            ax.yaxis.set_ticks_position('none')
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
        
    fig.tight_layout()
    return fig, axs
        





"""
Evaluation utility functions.
The original releas can be found here:https://github.com/qiaochen/VeloAE/
"""
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import cosine_similarity


def summary_scores(all_scores):
    """Summarize group scores.
    
    Args:
        all_scores (dict{str,list}): {group name: score list of individual cells}.
    
    Returns:
        dict{str,float}: Group-wise aggregation scores.
        float: score aggregated on all samples
        
    """
    sep_scores = {k:np.mean(s) for k, s in all_scores.items() if s }
    overal_agg = np.mean([s for k, s in sep_scores.items() if s])
    return sep_scores, overal_agg


def keep_type(adata, nodes, target, k_cluster):
    """Select cells of targeted type
    
    Args:
        adata (Anndata): Anndata object.
        nodes (list): Indexes for cells
        target (str): Cluster name.
        k_cluster (str): Cluster key in adata.obs dataframe

    Returns:
        list: Selected cells.

    """
    return nodes[adata.obs[k_cluster][nodes].values == target]


def cross_boundary_scvelo_probs(adata, k_cluster, cluster_edges, k_trans_g, return_raw=False):
    """Compute Cross-Boundary Confidence Score (A->B).
    
    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        cluster_edges (list of tuples("A", "B")): pairs of clusters has transition direction A->B
        k_trans_g (str): key to the transition graph computed using velocity.
        return_raw (bool): return aggregated or raw scores.
        
    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.
        
    """
    
    scores = {}
    all_scores = {}
    for u, v in cluster_edges:
        sel = adata.obs[k_cluster] == u
        nbs = adata.uns['neighbors']['indices'][sel]
        boundary_nodes = map(lambda nodes:keep_type(adata, nodes, v, k_cluster), nbs)
        type_score = [trans_probs.toarray()[:, nodes].mean() 
                      for trans_probs, nodes in zip(adata.uns[k_trans_g][sel], boundary_nodes) 
                      if len(nodes) > 0]
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
    if return_raw:
        return all_scores    
    return scores, np.mean([sc for sc in scores.values()])


def cross_boundary_coh(adata, k_cluster, k_velocity, cluster_edges, return_raw=False):
    """Cross-Boundary Velocity Coherence Score (A->B).
    
    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_velocity (str): key to the velocity matrix in adata.obsm.
        cluster_edges (list of tuples("A", "B")): pairs of clusters has transition direction A->B
        return_raw (bool): return aggregated or raw scores.
        
    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.
        
    """
    scores = {}
    all_scores = {}
    for u, v in cluster_edges:
        sel = adata.obs[k_cluster] == u
        nbs = adata.uns['neighbors']['indices'][sel]
        boundary_nodes = map(lambda nodes:keep_type(adata, nodes, v, k_cluster), nbs)
        
        velocities = adata.layers[k_velocity]
        v_us = velocities[sel]
        type_score = [cosine_similarity(v_us[[ith]], velocities[nodes]).mean()
                      for ith, nodes in enumerate(boundary_nodes) 
                      if len(nodes) > 0]
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
    
    if return_raw:
        return all_scores
    
    return scores, np.mean([sc for sc in scores.values()])

def in_cluster_scvelo_coh(adata, k_cluster, k_confidence, return_raw=False):
    """In-Cluster Confidence Score.
    
    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_confidence (str): key to the column of cell velocity confidence in adata.obs.
        return_raw (bool): return aggregated or raw scores.
        
    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.
    
    """
    clusters = np.unique(adata.obs[k_cluster])
    scores = {}
    all_scores = {}
    
    for cat in clusters:
        sel = adata.obs[k_cluster] == cat
        type_score = adata.obs[k_confidence][sel].values.tolist()
        scores[cat] = np.mean(type_score)
        all_scores[cat] = type_score
        
    if return_raw:
        return all_scores
    
    return scores, np.mean([s for _, s in scores.items()])

def cross_boundary_correctness(adata, 
                               k_cluster, 
                               k_velocity, 
                               cluster_edges, 
                               return_raw=False, 
                               x_emb="X_umap"):
    """Cross-Boundary Direction Correctness Score (A->B)
    
    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_velocity (str): key to the velocity matrix in adata.obsm.
        cluster_edges (list of tuples("A", "B")): pairs of clusters has transition direction A->B
        return_raw (bool): return aggregated or raw scores.
        x_emb (str): key to x embedding for visualization.
        
    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.
        
    """
    scores = {}
    all_scores = {}
    
    if x_emb == "X_umap":
        v_emb = adata.obsm['{}_umap'.format(k_velocity)]
    else:
        v_emb = adata.obsm[[key for key in adata.obsm if key.startswith(k_velocity)][0]]
    
    x_emb = adata.obsm[x_emb]
        
    for u, v in cluster_edges:
        sel = adata.obs[k_cluster] == u
        nbs = adata.uns['neighbors']['indices'][sel] # [n * 30]
        
        boundary_nodes = map(lambda nodes:keep_type(adata, nodes, v, k_cluster), nbs)
        x_points = x_emb[sel]
        x_velocities = v_emb[sel]
        
        type_score = []
        for x_pos, x_vel, nodes in zip(x_points, x_velocities, boundary_nodes):
            if len(nodes) == 0: continue

            position_dif = x_emb[nodes] - x_pos
            dir_scores = cosine_similarity(position_dif, x_vel.reshape(1,-1)).flatten()
            type_score.append(np.mean(dir_scores))
        
        scores[(u, v)] = np.mean(type_score)
        all_scores[(u, v)] = type_score
        
    if return_raw:
        return all_scores 
    
    return scores, np.mean([sc for sc in scores.values()])


def inner_cluster_coh(adata, k_cluster, k_velocity, return_raw=False):
    """In-cluster Coherence Score.
    
    Args:
        adata (Anndata): Anndata object.
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_velocity (str): key to the velocity matrix in adata.layers.
        return_raw (bool): return aggregated or raw scores.
        
    Returns:
        dict: all_scores indexed by cluster_edges
        or
        dict: mean scores indexed by cluster_edges
        float: averaged score over all cells.
        
    """
    clusters = np.unique(adata.obs[k_cluster])
    scores = {}
    all_scores = {}
    for cat in clusters:
        sel = adata.obs[k_cluster] == cat
        nbs = adata.uns['neighbors']['indices'][sel]
        same_cat_nodes = map(lambda nodes:keep_type(adata, nodes, cat, k_cluster), nbs)
        velocities = adata.layers[k_velocity]
        cat_vels = velocities[sel]
        cat_score = [cosine_similarity(cat_vels[[ith]], velocities[nodes]).mean() 
                     for ith, nodes in enumerate(same_cat_nodes) 
                     if len(nodes) > 0]
        all_scores[cat] = cat_score
        scores[cat] = np.mean(cat_score)
    
    if return_raw:
        return all_scores
    
    return scores, np.mean([sc for sc in scores.values()])


def evaluate(adata, cluster_edges, k_cluster, k_velocity, x_emb="X_umap", verbose=True):
    """Evaluate velocity estimation results using 5 metrics.
    
    Args:
        adata (Anndata): Anndata object.
        cluster_edges (list of tuples("A", "B")): pairs of clusters has transition direction A->B
        k_cluster (str): key to the cluster column in adata.obs DataFrame.
        k_velocity (str): key to the velocity matrix in adata.obsm.
        x_emb (str): key to x embedding for visualization.
        
    Returns:
        dict: aggregated metric scores.
    
    """
    
    trans_probs = cross_boundary_scvelo_probs(adata, k_cluster, cluster_edges, "{}_graph".format(k_velocity), True)
    crs_bdr_coh = cross_boundary_coh(adata, k_cluster, k_velocity, cluster_edges, True)
    crs_bdr_crc = cross_boundary_correctness(adata, k_cluster, k_velocity, cluster_edges, True, x_emb)
    ic_coh = inner_cluster_coh(adata, k_cluster, k_velocity, True)
    ic_scvelo_coh = in_cluster_scvelo_coh(adata, k_cluster, "{}_confidence".format(k_velocity), True)
    
    if verbose:
        print("# Cross-Boundary Transition Score (A->B)\n{}\nTotal Mean: {}".format(*summary_scores(trans_probs)))
        print("# Cross-Boundary Velocity Coherence (A->B)\n{}\nTotal Mean: {}".format(*summary_scores(crs_bdr_coh)))
        print("# Cross-Boundary Direction Correctness (A->B)\n{}\nTotal Mean: {}".format(*summary_scores(crs_bdr_crc)))
        print("# In-cluster Coherence\n{}\nTotal Mean: {}".format(*summary_scores(ic_coh)))
        print("# In-cluster Confidence\n{}\nTotal Mean: {}".format(*summary_scores(ic_scvelo_coh)))
    
    return {
        "Cross-Boundary Transition Score (A->B)": trans_probs,
        "Cross-Boundary Velocity Coherence (A->B)": crs_bdr_coh,
        "Cross-Boundary Direction Correctness (A->B)": crs_bdr_crc,
        "In-cluster Coherence": ic_coh,
        "In-cluster Confidence": ic_scvelo_coh
       }

def get_genescore(adata, nz_threshold=0.1, w_threshold=90, get_model=False,
                 scorenorm=True):
    model = np.zeros(adata.n_vars).astype(str)
    on_ = np.zeros(adata.n_vars)
    off_ = np.zeros(adata.n_vars)
    m1_ = np.zeros(adata.n_vars)
    m2_ = np.zeros(adata.n_vars)
    
    for i, g in enumerate(adata.var_names):
        c = adata[:, g].layers['c'].ravel()
        u = adata[:, g].layers['u'].ravel()
        s = adata[:, g].layers['s'].ravel()
        
        vc = adata[:, g].layers['velo_c'].ravel()
        vs = adata[:, g].layers['velo_s'].ravel()
        vu = adata[:, g].layers['velo_u'].ravel()
        
        w_non_zero = ((c >= nz_threshold * np.max(c)) &
                        (u >= nz_threshold * np.max(u)) &
                        (s >= nz_threshold * np.max(s)))
        u_non_zero = u[w_non_zero]
        s_non_zero = s[w_non_zero]
        wu = u >= np.percentile(u_non_zero, w_threshold)
        ws = s >= np.percentile(s_non_zero, w_threshold)
        
        #ww = wu|ws&down
        ww = wu|ws
        ww |= (vu * vs >= 0)
        
        on_[i] = np.mean((vu[w_non_zero]>0)&(vs[w_non_zero]>0))
        off_[i] = np.mean((vu[w_non_zero]<0)&(vs[w_non_zero]<0))
        m1_[i] = np.mean((vc[ww]<0)&(vu[ww]>0))
        m2_[i] = np.mean((vc[ww]>0)&(vu[ww]<0))

        if (np.mean(vu[ww]>0) > .7) & (np.mean(vs[ww]>0) >.7):
            model[i] = 'on'
        elif (np.mean(vu[ww]<0) > 0.8) & (np.mean(vs[ww]<0)>.8):
            model[i] = 'off'
        else:
            if np.mean(vc[ww]>0) > 0.5:
                model[i] = 'm2'
            else:
                model[i] = 'm1'


    model_df = pd.DataFrame({'on': on_, 'off': off_,
                         'm1': m1_, 'm2': m2_,},
                        index=adata.var_names)
    if scorenorm:
        scoresum = model_df['m1']+model_df['m2']
        model_df['m1'] /=scoresum
        model_df['m2'] /=scoresum
    
    if get_model:
        model_df['model'] = model
    
    return model_df
    
    


import pandas as pd
import numpy as np
from numpy.core.fromnumeric import size
import scipy
from sklearn.neighbors import NearestNeighbors
import matplotlib.pyplot as plt



def sampling_neighbors(gene_cus, step=(30,30), percentile=25):
    embed_len = gene_cus.shape[1]
    if embed_len != len(step):
        step = int(step[0]/(embed_len))
        step = np.repeat(step, embed_len)
    from scipy.stats import norm
    def gaussian_kernel(X, mu = 0, sigma=1):
        return np.exp(-(X - mu)**2 / (2*sigma**2)) / np.sqrt(2*np.pi*sigma**2)
    grs = []
    
    for dim_i in range(embed_len):
        m, M = np.min(gene_cus[:, dim_i]), np.max(gene_cus[:, dim_i])
        m = m - 0.025 * np.abs(M - m)
        M = M + 0.025 * np.abs(M - m)
        gr = np.linspace(m, M, step[dim_i])
        grs.append(gr)
    meshes_tuple = np.meshgrid(*grs)
    gridpoints_coordinates = np.vstack([i.flat for i in meshes_tuple]).T
    gridpoints_coordinates = gridpoints_coordinates + norm.rvs(loc=0, scale=0.15, size=gridpoints_coordinates.shape)
    
    np.random.seed(10) # set random seed
    
    nn = NearestNeighbors()

    neighbors_1 = min((gene_cus[:,:embed_len].shape[0]-1), 20)
    nn.fit(gene_cus[:,:embed_len])
    dist, ixs = nn.kneighbors(gridpoints_coordinates, neighbors_1)

    ix_choice = ixs[:,0].flat[:]
    ix_choice = np.unique(ix_choice)

    nn = NearestNeighbors()

    neighbors_2 = min((gene_cus.shape[0]-1), 20)
    nn.fit(gene_cus[:,0:embed_len])
    dist, ixs = nn.kneighbors(gene_cus[ix_choice, 0:embed_len], neighbors_2)
    
    density_estimate = gaussian_kernel(dist, mu=0, sigma=0.5).sum(1)
    bool_density = density_estimate > np.percentile(density_estimate, percentile)
    ix_choice = ix_choice[bool_density]
    return(ix_choice)

def sampling_inverse(gene_cus,target_amount=500):
    embed_len = gene_cus.shape[1]
    values = np.vstack([embed_len[:,i] for i in range(embed_len)])
    
    kernel = scipy.stats.gaussian_kde(values)
    p = kernel(values)
    # p2 = (1/p)/sum(1/p)
    p2 = (1/p)/sum(1/p)
    idx = np.arange(values.shape[1])
    r = scipy.stats.rv_discrete(values=(idx, p2))
    idx_choice = r.rvs(size=target_amount)
    return(idx_choice)

def sampling_circle(gene_cus,target_amount=500):
    embed_len = gene_cus.shape[1]
    values = np.vstack([embed_len[:,i] for i in range(embed_len)])
    
    kernel = scipy.stats.gaussian_kde(values)
    p = kernel(values)
    idx = np.arange(values.shape[1])
    tmp_p = np.square((1-(p/(max(p)))**2))+0.0001
    p2 = tmp_p/sum(tmp_p)
    r = scipy.stats.rv_discrete(values=(idx, p2))
    idx_choice = r.rvs(size=target_amount)
    return(idx_choice)

def sampling_random(gene_cus, target_amount=500):
    idx = np.random.choice(gene_cus.shape[0], size = target_amount, replace=False)
    return(idx)
    


def sampling_embedding(embedding,
                       para,
                    target_amount=500,
                    step=(30,30),
                      percentile=25):

    '''
    Guangyu
    '''
    if para == 'neighbors':
        idx = sampling_neighbors(embedding,step, percentile)
    elif para == 'inverse':
        #print('inverse')
        idx = sampling_inverse(embedding,target_amount)
    elif para == 'circle':
        idx = sampling_circle(embedding,target_amount)
    elif para == 'random':
        # print('random')
        idx = sampling_random(embedding,target_amount)
    else:
        print('downsampling_method is neighbors/inverse/circle/random')
    return(idx)


def adata_to_detail(data, para, gene):
    '''
    convert adata to detail format
    data: an anndata
    para: the varable name of unsplice, splice, and gene name
    para = ['Mc', 'Mu', 'Ms']
    '''
    data2 = data[:, data.var.index.isin([gene])].copy()
    chromosome = data2.layers[para[0]][:,0].copy().astype(np.float32)
    unsplice = data2.layers[para[1]][:,0].copy().astype(np.float32)
    splice = data2.layers[para[2]][:,0].copy().astype(np.float32)
    detail = pd.DataFrame({'gene_name':gene, 'chromosome': chromosome, 
                           'unsplice':unsplice, 'splice':splice})
    return(detail)

def downsampling_embedding(c, u, s, embedding,
                           para,
                           target_amount,
                           step,
                           n_neighbors,
                           embeddingnames=None,
                           expression_scale=None,
                           projection_neighbor_choice='embedding',
                          pca_n_components=None,
                           umap_n=None,
                           umap_n_components=None):
   
        
    
    if step is not None:
        idx_downSampling_embedding = sampling_embedding(embedding,
                    para=para,
                    target_amount=target_amount,
                    step=step)
    else:
        idx_downSampling_embedding=range(0,embedding.shape[0]) # all cells

    
    def transfer(c, u, s, expression_scale):
        

        if expression_scale=='log':
            c = np.log(c + 0.000001)
            u = np.log(u + 0.000001)
            s = np.log(s + 0.000001)
            
        elif expression_scale=='2power':
            c = 2**(c)
            u = 2**(u)
            s = 2**(s)
    
        elif expression_scale=='power10':
            c = c**10
            u = u**10
            s = s**10
            
        return (c, u, s)

    c, u, s =transfer(c, u, s, expression_scale)
    

    if projection_neighbor_choice=='gene':
        embedding_downsampling = s[idx_downSampling_embedding]
        
    elif projection_neighbor_choice=='pca': # not use
        from sklearn.decomposition import PCA
        #print('using pca projection_neighbor_choice')
        embedding_downsampling_0 = mdata.mod[rna_mod].layers['Ms'].iloc[idx_downSampling_embedding]
        pca=PCA(n_components=pca_n_components)
        pca.fit(embedding_downsampling_0)
        embedding_downsampling = pca.transform(embedding_downsampling_0)[:,range(pca_n_components)]
        
    elif projection_neighbor_choice=='pca_norm':
        from sklearn.decomposition import PCA
        #print('pca_norm')
        embedding_downsampling_0 = s[idx_downSampling_embedding]
        
        pca=PCA(n_components=pca_n_components)
        pca.fit(embedding_downsampling_0)
        embedding_downsampling_trans = pca.transform(embedding_downsampling_0)[:,range(pca_n_components)]
        embedding_downsampling_trans_norm=(embedding_downsampling_trans - embedding_downsampling_trans.min(0)) / embedding_downsampling_trans.ptp(0)#normalize
        embedding_downsampling_trans_norm_mult10=embedding_downsampling_trans_norm*10 #optional
        embedding_downsampling=embedding_downsampling_trans_norm_mult10**5 # optional
        
    elif projection_neighbor_choice=='embedding':
        embedding_downsampling = embedding[idx_downSampling_embedding]

    elif projection_neighbor_choice =='umap':
        import umap
        embedding_downsampling_0 = s[idx_downSampling_embedding]
        
        def get_umap(df,n_neighbors=umap_n, min_dist=0.1, n_components=umap_n_components, metric='euclidean'): 
            fit = umap.UMAP(
                n_neighbors=n_neighbors,
                min_dist=min_dist,
                n_components=n_components,
                metric=metric
            )
            embed = fit.fit_transform(df);
            return(embed)
        embedding_downsampling=get_umap(embedding_downsampling_0)

    n_neighbors = min(int((embedding_downsampling.shape[0])/4), n_neighbors)
    if n_neighbors==0:
        n_neighbors=1
    nn = NearestNeighbors(n_neighbors=n_neighbors)
    nn.fit(embedding_downsampling) 
    embedding_knn = nn.kneighbors_graph(mode="connectivity")
    return(embedding_downsampling, idx_downSampling_embedding, embedding_knn)

def downsampling(data_df, gene_list, downsampling_ixs):
    '''
    Guangyu
    '''
    data_df_downsampled=pd.DataFrame()
    for gene in gene_list:
        data_df_one_gene=data_df[data_df['gene_name']==gene]
        data_df_one_gene_downsampled = data_df_one_gene.iloc[downsampling_ixs]
        data_df_downsampled=data_df_downsampled.append(data_df_one_gene_downsampled)
    return(data_df_downsampled)
